from __future__ import annotations

import logging
import time
import typing as t
import uuid

from globus_sdk import client, exc, paging, response
from globus_sdk._internal import guards
from globus_sdk._internal.remarshal import commajoin
from globus_sdk._internal.type_definitions import DateLike, IntLike
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.response import IterableJSONAPIResponse
from globus_sdk.scopes import GCSCollectionScopes, Scope, TransferScopes
from globus_sdk.transport import RetryConfig

from .data import CreateTunnelData, DeleteData, TransferData
from .errors import TransferAPIError
from .response import IterableTransferResponse
from .transport import TRANSFER_DEFAULT_RETRY_CHECKS

log = logging.getLogger(__name__)

TransferFilterDict = t.Dict[str, t.Union[str, t.List[str]]]


def _datelike_to_str(x: DateLike) -> str:
    return x if isinstance(x, str) else x.isoformat(timespec="seconds")


def _format_completion_time(
    x: str | tuple[DateLike, DateLike] | MissingType,
) -> str | MissingType:
    if x is MISSING:
        return MISSING
    elif isinstance(x, str):
        return x
    else:
        start_t, end_t = x
        start_t, end_t = _datelike_to_str(start_t), _datelike_to_str(end_t)
        return f"{start_t},{end_t}"


@t.overload
def _format_filter_item(x: str | TransferFilterDict) -> str: ...


@t.overload
def _format_filter_item(x: MissingType) -> MissingType: ...


def _format_filter_item(x: str | TransferFilterDict | MissingType) -> str | MissingType:
    if x is MISSING:
        return MISSING
    elif isinstance(x, str):
        return x
    return "/".join(f"{k}:{commajoin(v)}" for k, v in x.items())


def _format_filter(
    x: str | TransferFilterDict | list[str | TransferFilterDict] | MissingType,
) -> str | list[str] | MissingType:
    if isinstance(x, list):
        return [_format_filter_item(y) for y in x]
    return _format_filter_item(x)


def _get_page_size(paged_result: IterableTransferResponse) -> int:
    return len(paged_result["DATA"])


class TransferClient(client.BaseClient):
    r"""
    Client for the
    `Globus Transfer API <https://docs.globus.org/api/transfer/>`_.

    .. sdk-sphinx-copy-params:: BaseClient

    This class provides helper methods for most common resources in the
    REST API, and basic ``get``, ``put``, ``post``, and ``delete`` methods
    from the base rest client that can be used to access any REST resource.

    Detailed documentation is available in the official REST API
    documentation, which is linked to from the method documentation. Methods
    that allow arbitrary keyword arguments will pass the extra arguments as
    query parameters.

    .. _transfer_filter_formatting:

    **Filter Formatting**

    Several methods of ``TransferClient`` take a ``filter`` parameter which can be a
    string, dict, or list of strings/dicts. When the filter given is a string, it is
    passed as a single unmodified param. When the filter given is a dict, it is
    formatted according to the below rules into a single param. When the filter is a
    list, each item of the list is parsed and passed as separate params.

    dict parsing rules:

    - each (key, value) pair in the dict is a clause in the resulting filter string
    - clauses are each formatted to ``key:value``
    - when the value is a list, it is comma-separated, as in ``key:value1,value2``
    - clauses are separated with slashes, as in ``key1:value1/key2:value2``

    The corresponding external API documentation describes, in detail, the supported
    filter clauses for each method which uses the ``filter`` parameter.
    Generally, speaking, filter clauses documented as ``string list`` can be passed as
    lists to a filter dict, while string, date, and numeric filters should be passed as
    strings.

    **Paginated Calls**

    Methods which support pagination can be called as paginated or unpaginated methods.
    If the method name is ``TransferClient.foo``, the paginated version is
    ``TransferClient.paginated.foo``.
    Using ``TransferClient.endpoint_search`` as an example::

        from globus_sdk import TransferClient
        tc = TransferClient(...)

        # this is the unpaginated version
        for x in tc.endpoint_search("tutorial"):
            print("Endpoint ID: {}".format(x["id"]))

        # this is the paginated version
        for page in tc.paginated.endpoint_search("testdata"):
            for x in page:
                print("Endpoint ID: {}".format(x["id"]))

    .. automethodlist:: globus_sdk.TransferClient
    """

    service_name = "transfer"
    error_class = TransferAPIError
    scopes = TransferScopes
    default_scope_requirements = [TransferScopes.all]

    def _register_standard_retry_checks(self, retry_config: RetryConfig) -> None:
        """Override the default retry checks."""
        retry_config.checks.register_many_checks(TRANSFER_DEFAULT_RETRY_CHECKS)

    def add_app_data_access_scope(
        self, collection_ids: uuid.UUID | str | t.Iterable[uuid.UUID | str]
    ) -> TransferClient:
        """
        Add a dependent ``data_access`` scope for one or more given ``collection_ids``
        to this client's ``GlobusApp``. Useful for resolving ``ConsentRequired`` errors
        when using standard Globus Connect Server mapped collections.

        .. warning::

            This method must only be used on ``collection_ids`` for non-High-Assurance
            GCS Mapped Collections.

            Use on other collection types, e.g., on GCP Mapped Collections or any form
            of Guest Collection, will result in "Unknown Scope" errors during the login
            flow.

        Returns ``self`` for chaining.

        Raises ``GlobusSDKUsageError`` if this client was not initialized with an app.

        :param collection_ids: a collection ID or an iterable of IDs.

        .. tab-set::

            .. tab-item:: Example Usage

                Usage for ``ls`` or a similar operation which points at a single
                collection:

                .. code-block:: python

                    app = UserApp("myapp", client_id=NATIVE_APP_CLIENT_ID)
                    client = TransferClient(app=app).add_app_data_access_scope(COLLECTION_ID)

                    res = client.operation_ls(COLLECTION_ID)

                Usage for ``submit_transfer`` or a similar operation which points at
                multiple collections:

                .. code-block:: python

                    app = UserApp("myapp", client_id=NATIVE_APP_CLIENT_ID)
                    client = TransferClient(app=app).add_app_data_access_scope(
                        (COLLECTION_ID_1, COLLECTION_ID_2)
                    )

                    transfer_data = TransferData(
                        source_endpoint=COLLECTION_ID_1,
                        destination_endpoint=COLLECTION_ID_2,
                    )
                    transfer_data.add_item("/foo", "bar/baz")
                    ...
                    res = client.submit_transfer({})
        """  # noqa: E501
        if isinstance(collection_ids, (str, uuid.UUID)):
            guards.validators.uuidlike("collection_ids", collection_ids)
            # wrap the collection_ids input in a list for consistent iteration below
            collection_ids_ = [collection_ids]
        else:
            # copy to a list so that ephemeral iterables can be iterated multiple times
            collection_ids_ = list(collection_ids)
            for i, c in enumerate(collection_ids_):
                guards.validators.uuidlike(f"collection_ids[{i}]", c)

        scope = TransferScopes.all
        dependencies: list[Scope] = []
        for coll_id in collection_ids_:
            data_access_scope = GCSCollectionScopes(
                str(coll_id)
            ).data_access.with_optional(True)
            dependencies.append(data_access_scope)
        scope = scope.with_dependencies(dependencies)
        self.add_app_scope(scope)
        return self

    # Convenience methods, providing more pythonic access to common REST
    # resources

    #
    # Endpoint Management
    #

    def get_endpoint(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: ID of endpoint to lookup
        :param query_params: Any additional parameters will be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    endpoint = tc.get_endpoint(endpoint_id)
                    print("Endpoint name:", endpoint["display_name"] or endpoint["canonical_name"])

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>``

                .. extdoclink:: Get Endpoint or Collection by ID
                    :ref: transfer/endpoints_and_collections/#get_endpoint_or_collection_by_id
        """  # noqa: E501
        log.debug(f"TransferClient.get_endpoint({endpoint_id})")
        return self.get(f"/v0.10/endpoint/{endpoint_id}", query_params=query_params)

    def update_endpoint(
        self,
        endpoint_id: uuid.UUID | str,
        data: dict[str, t.Any],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: ID of endpoint to lookup
        :param data: A partial endpoint document with fields to update
        :param query_params: Any additional parameters will be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    epup = {
                        "display_name": "My New Endpoint Name",
                        "description": "Better Description",
                    }
                    update_result = tc.update_endpoint(endpoint_id, epup)

            .. tab-item:: API Info

                ``PUT /endpoint/<endpoint_id>``

                .. extdoclink:: Update Globus Connect Personal collection by id
                    :ref: transfer/gcp_management/#update_collection_by_id
        """  # noqa: E501
        log.debug(f"TransferClient.update_endpoint({endpoint_id}, ...)")
        return self.put(
            f"/v0.10/endpoint/{endpoint_id}", data=data, query_params=query_params
        )

    def set_subscription_id(
        self,
        collection_id: uuid.UUID | str,
        subscription_id: uuid.UUID | str | t.Literal["DEFAULT"] | None,
    ) -> response.GlobusHTTPResponse:
        """
        Set the ``subscription_id`` on a mapped collection.
        This is primarily used for subscription management on Globus Connect Personal.

        This operation requires membership in a Globus subscription group and
        has authorization requirements which depend upon the caller's roles on
        the subscription group and the endpoint or collection.

        Guest Collections inherit the subscriptions of their Mapped Collections and
        cannot have a ``subscription_id`` directly set in this way.

        :param collection_id: The collection ID which is having its subscription set.
        :param subscription_id: The ID of the subscription to assign, the special
            string ``"DEFAULT"``, or ``None``.

        .. note::

            Setting ``subscription_id="DEFAULT"`` results in the service choosing your
            subscription ID, but requires that you only have one subscription.
            If you have multiple subscriptions, using ``"DEFAULT"`` will result in an
            error.

            Setting ``subscription_id=None`` clears any existing subscription from the
            collection.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    import globus_sdk

                    MY_SUBSCRIPTION_ID = "..."
                    LOCAL_GCP = globus_sdk.LocalGlobusConnectPersonal()

                    tc = globus_sdk.TransferClient(...)
                    tc.endpoint_set_subscription_id(
                        LOCAL_GCP.endpoint_id,
                        MY_SUBSCRIPTION_ID,
                    )

            .. tab-item:: API Info

                ``PUT /endpoint/<collection_id>/subscription``

                .. extdoclink:: Associate a Globus Connect Personal Mapped Collection with a Subscription
                    :ref: transfer/gcp_management/#associate_collection_subscription
        """  # noqa: E501
        return self.put(
            f"/v0.10/endpoint/{collection_id}/subscription",
            data={"subscription_id": subscription_id},
        )

    def set_subscription_admin_verified(
        self,
        collection_id: uuid.UUID | str,
        subscription_admin_verified: bool,
    ) -> response.GlobusHTTPResponse:
        """
        Sets the value of ``subscription_admin_verified`` on a Globus Connect Personal
        mapped collection. A value of ``True`` grants verified status, and a
        value of ``False`` revokes verified status.

        This operation requires membership in a Globus subscription group and
        has authorization requirements which depend upon the caller's roles on
        the subscription group and the collection.

        Subscription administrators can grant or revoke verification on a collection
        that is associated with their subscription without needing an administrator
        role on the collection itself.

        Users with the administrator effective role on the collection can revoke
        verification on a collection, but must still be a subscription administrator
        to grant verification.

        :param collection_id: The collection ID which is having its subscription set.
        :param subscription_admin_verified: The verification status of the collection
            expressed as a Boolean type.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    import globus_sdk

                    LOCAL_GCP = globus_sdk.LocalGlobusConnectPersonal()

                    tc = globus_sdk.TransferClient(...)
                    tc.endpoint_set_subscription_id(
                        LOCAL_GCP.endpoint_id,
                        True,
                    )

            .. tab-item:: API Info

                ``PUT /endpoint/<collection_id>/subscription_admin_verified``

                .. extdoclink:: Set Subscription Admin Verified
                    :ref: transfer/gcp_management/#set_subscription_admin_verified
        """  # noqa: E501
        return self.put(
            f"/v0.10/endpoint/{collection_id}/subscription_admin_verified",
            data={"subscription_admin_verified": subscription_admin_verified},
        )

    def delete_endpoint(
        self, endpoint_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: ID of endpoint to delete

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    delete_result = tc.delete_endpoint(endpoint_id)

            .. tab-item:: API Info

                ``DELETE /endpoint/<endpoint_id>``

                .. extdoclink:: Delete Globus Connect Personal collection by id
                    :ref: transfer/gcp_management/#delete_collection_by_id
        """
        log.debug(f"TransferClient.delete_endpoint({endpoint_id})")
        return self.delete(f"/v0.10/endpoint/{endpoint_id}")

    @paging.has_paginator(
        paging.HasNextPaginator,
        items_key="DATA",
        get_page_size=_get_page_size,
        max_total_results=1000,
        page_size=100,
    )
    def endpoint_search(
        self,
        filter_fulltext: str | MissingType = MISSING,
        *,
        filter_scope: str | MissingType = MISSING,
        filter_owner_id: str | MissingType = MISSING,
        filter_host_endpoint: uuid.UUID | str | MissingType = MISSING,
        filter_non_functional: bool | MissingType = MISSING,
        filter_entity_type: (
            t.Literal[
                "GCP_mapped_collection",
                "GCP_guest_collection",
                "GCSv5_endpoint",
                "GCSv5_mapped_collection",
                "GCSv5_guest_collection",
            ]
            | MissingType
        ) = MISSING,
        limit: int | MissingType = MISSING,
        offset: int | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        r"""
        :param filter_fulltext: The string to use in a full text search on endpoints.
            Effectively, the "search query" which is being requested. May be omitted
            with specific ``filter_scope`` values.
        :param filter_scope: A "scope" within which to search for endpoints. This must
            be one of the limited and known names known to the service, which can be
            found documented in the **External Documentation** below. Defaults to
            searching all endpoints (in which case ``filter_fulltext`` is required)
        :param filter_owner_id: Limit search to endpoints owned by the specified Globus
            Auth identity. Conflicts with scopes 'my-endpoints', 'my-gcp-endpoints', and
            'shared-by-me'.
        :param filter_host_endpoint: Limit search to endpoints hosted by the specified
            endpoint. May cause BadRequest or PermissionDenied errors if the endpoint ID
            given is not valid for this operation.
        :param filter_non_functional: Limit search to endpoints which have the
            'non_functional' flag set to True or False. Mutually exclusive with
            ``filter_entity_type``.
        :param filter_entity_type: Limit search to endpoints or collections of a
            specified entity type. Mutually exclusive with ``filter_non_functional``.
        :param limit: limit the number of results
        :param offset: offset used in paging
        :param query_params: Any additional parameters will be passed through
            as query params.

        It is important to be aware that the Endpoint Search API limits
        you to 1000 results for any search query.

        .. tab-set::

            .. tab-item:: Example Usage

                Search for a given string as a fulltext search:

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    for ep in tc.endpoint_search("String to search for!"):
                        print(ep["display_name"])

                Search for a given string, but only on endpoints that you own:

                .. code-block:: python

                    for ep in tc.endpoint_search("foo", filter_scope="my-endpoints"):
                        print(f"{ep['display_name']} has ID {ep['id']}")

            .. tab-item:: Paginated Usage

                .. paginatedusage:: endpoint_search

            .. tab-item:: API Info

                .. parsed-literal::

                    GET /endpoint_search\
                    ?filter_fulltext=<filter_fulltext>&filter_scope=<filter_scope>

                .. extdoclink:: Endpoint and Collection Search
                    :ref: transfer/endpoint_and_collection_search
        """  # noqa: E501
        query_params = {
            "filter_scope": filter_scope,
            "filter_fulltext": filter_fulltext,
            "filter_owner_id": filter_owner_id,
            "filter_host_endpoint": filter_host_endpoint,
            "filter_non_functional": (
                1
                if filter_non_functional
                else (
                    0
                    if isinstance(filter_non_functional, bool)
                    else filter_non_functional
                )
            ),
            "filter_entity_type": filter_entity_type,
            "limit": limit,
            "offset": offset,
            **(query_params or {}),
        }
        log.debug(f"TransferClient.endpoint_search({query_params})")
        return IterableTransferResponse(
            self.get("/v0.10/endpoint_search", query_params=query_params)
        )

    def my_effective_pause_rule_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        :param endpoint_id: the endpoint on which the current user's effective pause
            rules are fetched
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/my_effective_pause_rule_list``

                .. extdoclink:: Get my effective collection pause rules
                    :ref: transfer/endpoints_and_collections/#get_collection_pause_rules
        """
        log.debug(f"TransferClient.my_effective_pause_rule_list({endpoint_id}, ...)")
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint/{endpoint_id}/my_effective_pause_rule_list",
                query_params=query_params,
            )
        )

    # Shared Endpoints

    def my_shared_endpoint_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        :param endpoint_id: the host endpoint whose shares are listed
        :param query_params: Additional passthrough query parameters

        Get a list of shared endpoints for which the user has ``administrator`` or
        ``access_manager`` on a given host endpoint.

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/my_shared_endpoint_list``

                .. extdoclink:: Get my guest collection list
                    :ref: transfer/endpoints_and_collections/#get_my_guest_collection_list
        """  # noqa: E501
        log.debug(f"TransferClient.my_shared_endpoint_list({endpoint_id}, ...)")
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint/{endpoint_id}/my_shared_endpoint_list",
                query_params=query_params,
            )
        )

    @paging.has_paginator(paging.NextTokenPaginator, items_key="shared_endpoints")
    def get_shared_endpoint_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        max_results: int | MissingType = MISSING,
        next_token: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        :param endpoint_id: the host endpoint whose shares are listed
        :param max_results: cap to the number of results
        :param next_token: token used for paging
        :param query_params: Any additional parameters to be passed through
            as query params.

        Get a list of all shared endpoints on a given host endpoint.

        .. tab-set::

            .. tab-item:: Paginated Usage

                .. paginatedusage:: get_shared_endpoint_list

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/shared_endpoint_list``

                .. extdoclink:: Get guest collection list
                    :ref: transfer/endpoints_and_collections/get_guest_collection_list
        """
        log.debug(f"TransferClient.get_shared_endpoint_list({endpoint_id}, ...)")
        query_params = {
            "max_results": (
                str(max_results) if isinstance(max_results, int) else max_results
            ),
            "next_token": next_token,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint/{endpoint_id}/shared_endpoint_list",
                query_params=query_params,
            ),
            iter_key="shared_endpoints",
        )

    def create_shared_endpoint(
        self, data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        :param data: A python dict representation of a ``shared_endpoint`` document

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    shared_ep_data = {
                        "DATA_TYPE": "shared_endpoint",
                        "host_endpoint": host_endpoint_id,
                        "host_path": host_path,
                        "display_name": display_name,
                        # optionally specify additional endpoint fields
                        "description": "my test share",
                    }
                    create_result = tc.create_shared_endpoint(shared_ep_data)
                    endpoint_id = create_result["id"]

            .. tab-item:: API Info

                ``POST /shared_endpoint``

                .. extdoclink:: Create guest collection
                    :ref: transfer/gcp_management/#create_guest_collection
        """
        log.debug("TransferClient.create_shared_endpoint(...)")
        return self.post("/v0.10/shared_endpoint", data=data)

    # Endpoint servers

    def endpoint_server_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        :param endpoint_id: The endpoint whose servers are being listed
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/server_list``

                .. extdoclink:: Get endpoint or collection server list
                    :ref: transfer/endpoints_and_collections/#get_endpoint_or_collection_server_list
        """  # noqa: E501
        log.debug(f"TransferClient.endpoint_server_list({endpoint_id}, ...)")
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint/{endpoint_id}/server_list", query_params=query_params
            )
        )

    def get_endpoint_server(
        self,
        endpoint_id: uuid.UUID | str,
        server_id: IntLike,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The endpoint under which the server is registered
        :param server_id: The ID of the server
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/server/<server_id>``

                .. extdoclink:: Get server by id
                    :ref: transfer/endpoints_and_collections/#get_server_by_id
        """
        log.debug(
            "TransferClient.get_endpoint_server(%s, %s, ...)", endpoint_id, server_id
        )
        return self.get(
            f"/v0.10/endpoint/{endpoint_id}/server/{server_id}",
            query_params=query_params,
        )

    #
    # Roles
    #

    def endpoint_role_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        :param endpoint_id: The endpoint whose roles are being listed
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/role_list``

                .. extdoclink:: Get list of roles
                    :ref: transfer/roles/#role_list
        """
        log.debug(f"TransferClient.endpoint_role_list({endpoint_id}, ...)")
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint/{endpoint_id}/role_list", query_params=query_params
            )
        )

    def add_endpoint_role(
        self, endpoint_id: uuid.UUID | str, role_data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The endpoint on which the role is being added
        :param role_data: A role document for the new role

        .. tab-set::

            .. tab-item:: API Info

                ``POST /endpoint/<endpoint_id>/role``

                .. extdoclink:: Create Globus Connect Personal collection role
                    :ref: transfer/roles/#create_role
        """
        log.debug(f"TransferClient.add_endpoint_role({endpoint_id}, ...)")
        return self.post(f"/v0.10/endpoint/{endpoint_id}/role", data=role_data)

    def get_endpoint_role(
        self,
        endpoint_id: uuid.UUID | str,
        role_id: str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The endpoint on which the role applies
        :param role_id: The ID of the role
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/role/<role_id>``

                .. extdoclink:: Get role by id
                    :ref: transfer/roles/#get_role_by_id
        """
        log.debug(f"TransferClient.get_endpoint_role({endpoint_id}, {role_id}, ...)")
        return self.get(
            f"/v0.10/endpoint/{endpoint_id}/role/{role_id}", query_params=query_params
        )

    def delete_endpoint_role(
        self, endpoint_id: uuid.UUID | str, role_id: str
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The endpoint on which the role applies
        :param role_id: The ID of the role to delete

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /endpoint/<endpoint_id>/role/<role_id>``

                .. extdoclink:: Delete Globus Connect Personal collection role by id
                    :ref: transfer/roles/#delete_role_by_id
        """
        log.debug(f"TransferClient.delete_endpoint_role({endpoint_id}, {role_id})")
        return self.delete(f"/v0.10/endpoint/{endpoint_id}/role/{role_id}")

    #
    # ACLs
    #

    def endpoint_acl_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        :param endpoint_id: The endpoint whose ACLs are being listed
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/access_list``

                .. extdoclink:: Get list of access rules
                    :ref: transfer/acl/#rest_access_get_list
        """
        log.debug(f"TransferClient.endpoint_acl_list({endpoint_id}, ...)")
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint/{endpoint_id}/access_list", query_params=query_params
            )
        )

    def get_endpoint_acl_rule(
        self,
        endpoint_id: uuid.UUID | str,
        rule_id: str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The endpoint on which the access rule applies
        :param rule_id: The ID of the rule to fetch
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint/<endpoint_id>/access/<rule_id>``

                .. extdoclink:: Get access rule by ID
                    :ref: transfer/acl/#get_access_rule_by_id
        """
        log.debug(
            "TransferClient.get_endpoint_acl_rule(%s, %s, ...)", endpoint_id, rule_id
        )
        return self.get(
            f"/v0.10/endpoint/{endpoint_id}/access/{rule_id}", query_params=query_params
        )

    def add_endpoint_acl_rule(
        self, endpoint_id: uuid.UUID | str, rule_data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: ID of endpoint to which to add the acl
        :param rule_data: A python dict representation of an ``access`` document

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    rule_data = {
                        "DATA_TYPE": "access",
                        "principal_type": "identity",
                        "principal": identity_id,
                        "path": "/dataset1/",
                        "permissions": "rw",
                    }
                    result = tc.add_endpoint_acl_rule(endpoint_id, rule_data)
                    rule_id = result["access_id"]

                Note that if this rule is being created on a shared endpoint
                the "path" field is relative to the "host_path" of the shared endpoint.

            .. tab-item:: API Info

                ``POST /endpoint/<endpoint_id>/access``

                .. extdoclink:: Create access rule
                    :ref: transfer/acl/#rest_access_create
        """
        log.debug(f"TransferClient.add_endpoint_acl_rule({endpoint_id}, ...)")
        return self.post(f"/v0.10/endpoint/{endpoint_id}/access", data=rule_data)

    def update_endpoint_acl_rule(
        self,
        endpoint_id: uuid.UUID | str,
        rule_id: str,
        rule_data: dict[str, t.Any],
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The endpoint on which the access rule applies
        :param rule_id: The ID of the access rule to modify
        :param rule_data: A partial ``access`` document containing fields to update

        .. tab-set::

            .. tab-item:: API Info

                ``PUT /endpoint/<endpoint_id>/access/<rule_id>``

                .. extdoclink:: Update access rule
                    :ref: transfer/acl/#update_access_rule
        """
        log.debug(
            "TransferClient.update_endpoint_acl_rule(%s, %s, ...)",
            endpoint_id,
            rule_id,
        )
        return self.put(
            f"/v0.10/endpoint/{endpoint_id}/access/{rule_id}", data=rule_data
        )

    def delete_endpoint_acl_rule(
        self, endpoint_id: uuid.UUID | str, rule_id: str
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The endpoint on which the access rule applies
        :param rule_id: The ID of the access rule to remove

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /endpoint/<endpoint_id>/access/<rule_id>``

                .. extdoclink:: Delete access rule
                    :ref: transfer/acl/#delete_access_rule
        """
        log.debug(
            "TransferClient.delete_endpoint_acl_rule(%s, %s)", endpoint_id, rule_id
        )
        return self.delete(f"/v0.10/endpoint/{endpoint_id}/access/{rule_id}")

    #
    # Bookmarks
    #

    def bookmark_list(
        self, *, query_params: dict[str, t.Any] | None = None
    ) -> IterableTransferResponse:
        """
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /bookmark_list``

                .. extdoclink:: Get list of bookmarks
                    :ref: transfer/collection_bookmarks/#get_list_of_bookmarks
        """
        log.debug(f"TransferClient.bookmark_list({query_params})")
        return IterableTransferResponse(
            self.get("/v0.10/bookmark_list", query_params=query_params)
        )

    def create_bookmark(
        self, bookmark_data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        :param bookmark_data: A bookmark document for the bookmark to create

        .. tab-set::

            .. tab-item:: API Info

                ``POST /bookmark``

                .. extdoclink:: Create bookmark
                    :ref: transfer/collection_bookmarks/#create_bookmark
        """
        log.debug(f"TransferClient.create_bookmark({bookmark_data})")
        return self.post("/v0.10/bookmark", data=bookmark_data)

    def get_bookmark(
        self,
        bookmark_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param bookmark_id: The ID of the bookmark to lookup
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /bookmark/<bookmark_id>``

                .. extdoclink:: Get bookmark by ID
                    :ref: transfer/collection_bookmarks/#get_bookmark_by_id
        """
        log.debug(f"TransferClient.get_bookmark({bookmark_id})")
        return self.get(f"/v0.10/bookmark/{bookmark_id}", query_params=query_params)

    def update_bookmark(
        self, bookmark_id: uuid.UUID | str, bookmark_data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        :param bookmark_id: The ID of the bookmark to modify
        :param bookmark_data: A partial bookmark document with fields to update

        .. tab-set::

            .. tab-item:: API Info

                ``PUT /bookmark/<bookmark_id>``

                .. extdoclink:: Update bookmark
                    :ref: transfer/collection_bookmarks/#update_bookmark
        """
        log.debug(f"TransferClient.update_bookmark({bookmark_id})")
        return self.put(f"/v0.10/bookmark/{bookmark_id}", data=bookmark_data)

    def delete_bookmark(
        self, bookmark_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        :param bookmark_id: The ID of the bookmark to delete

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /bookmark/<bookmark_id>``

                .. extdoclink:: Delete bookmark by ID
                    :ref: transfer/collection_bookmarks/#delete_bookmark_by_id
        """
        log.debug(f"TransferClient.delete_bookmark({bookmark_id})")
        return self.delete(f"/v0.10/bookmark/{bookmark_id}")

    #
    # Synchronous Filesys Operations
    #

    def operation_ls(
        self,
        endpoint_id: uuid.UUID | str,
        path: str | MissingType = MISSING,
        *,
        show_hidden: bool | MissingType = MISSING,
        orderby: str | list[str] | MissingType = MISSING,
        limit: int | MissingType = MISSING,
        offset: int | MissingType = MISSING,
        # note: filter is a soft keyword in python, so using this name is okay
        # pylint: disable=redefined-builtin
        filter: (
            str | TransferFilterDict | list[str | TransferFilterDict] | MissingType
        ) = MISSING,
        local_user: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        :param endpoint_id: The ID of the endpoint on which to do a dir listing
        :param path: Path to a directory on the endpoint to list
        :param show_hidden: Show hidden files (names beginning in dot).
            Defaults to true.
        :param limit: Limit the number of results returned. Defaults to 100,000 ,
            which is also the maximum.
        :param offset: Offset into the result list, which can be used to page results.
        :param orderby: One or more order-by options. Each option is
            either a field name or a field name followed by a space and 'ASC' or 'DESC'
            for ascending or descending.
        :param filter: Only return file documents which match these filter clauses. For
            the filter syntax, see the **External Documentation** linked below. If a
            dict is supplied as the filter, it is formatted as a set of filter clauses.
            If a list is supplied, it is passed as multiple params.
            See :ref:`filter formatting <transfer_filter_formatting>` for details.
        :param local_user: Optional value passed to identity mapping specifying which
            local user account to map to. Only usable with Globus Connect Server v5
            mapped collections.
        :param query_params: Additional passthrough query parameters

        .. note::

            Pagination is not supported by the GridFTP protocol, and therefore
            limit+offset pagination will result in the Transfer service repeatedly
            fetching an entire directory listing from the server and filtering it before
            returning it to the client.

            For latency-sensitive applications, such usage may still be more efficient
            than asking for a very large directory listing as it reduces the size of the
            payload passed between the Transfer service and the client.

        .. tab-set::

            .. tab-item:: Example Usage

                List with a path:

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    for entry in tc.operation_ls(ep_id, path="/~/project1/"):
                        print(entry["name"], entry["type"])

                List with explicit ordering:

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    for entry in tc.operation_ls(ep_id, path="/~/project1/", orderby=["type", "name"]):
                        print(entry["name DESC"], entry["type"])

                List filtering to files modified before January 1, 2021. Note the use
                of an empty "start date" for the filter:

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    for entry in tc.operation_ls(
                        ep_id,
                        path="/~/project1/",
                        filter={"last_modified": ["", "2021-01-01"]},
                    ):
                        print(entry["name"], entry["type"])

            .. tab-item:: API Info

                ``GET /operation/endpoint/<endpoint_id>/ls``

                .. extdoclink:: List Directory Contents
                    :ref: transfer/file_operations/#list_directory_contents
        """  # noqa: E501
        query_params = {
            "path": path,
            "limit": limit,
            "offset": offset,
            "show_hidden": (
                1
                if show_hidden
                else 0 if isinstance(show_hidden, bool) else show_hidden
            ),
            "orderby": commajoin(orderby),
            "filter": _format_filter(filter),
            "local_user": local_user,
            **(query_params or {}),
        }
        log.debug(f"TransferClient.operation_ls({endpoint_id}, {query_params})")
        return IterableTransferResponse(
            self.get(
                f"/v0.10/operation/endpoint/{endpoint_id}/ls", query_params=query_params
            )
        )

    def operation_mkdir(
        self,
        endpoint_id: uuid.UUID | str,
        path: str,
        *,
        local_user: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The ID of the endpoint on which to create a directory
        :param path: Path to the new directory to create
        :param local_user: Optional value passed to identity mapping specifying which
            local user account to map to. Only usable with Globus Connect Server v5
            mapped collections.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    tc.operation_mkdir(ep_id, path="/~/newdir/")

            .. tab-item:: API Info

                ``POST /operation/endpoint/<endpoint_id>/mkdir``

                .. extdoclink:: Make Directory
                    :ref: transfer/file_operations/#make_directory
        """
        log.debug(
            "TransferClient.operation_mkdir({}, {}, {})".format(
                endpoint_id, path, query_params
            )
        )
        json_body = {
            "DATA_TYPE": "mkdir",
            "path": path,
            "local_user": local_user,
        }
        return self.post(
            f"/v0.10/operation/endpoint/{endpoint_id}/mkdir",
            data=json_body,
            query_params=query_params,
        )

    def operation_rename(
        self,
        endpoint_id: uuid.UUID | str,
        oldpath: str,
        newpath: str,
        *,
        local_user: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The ID of the endpoint on which to rename a file
        :param oldpath: Path to the old filename
        :param newpath: Path to the new filename
        :param local_user: Optional value passed to identity mapping specifying which
            local user account to map to. Only usable with Globus Connect Server v5
            mapped collections.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    tc.operation_rename(ep_id, oldpath="/~/file1.txt", newpath="/~/project1data.txt")

            .. tab-item:: API Info

                ``POST /operation/endpoint/<endpoint_id>/rename``

                .. extdoclink:: Rename
                    :ref: transfer/file_operations/#rename
        """  # noqa: E501
        log.debug(
            "TransferClient.operation_rename({}, {}, {}, {})".format(
                endpoint_id, oldpath, newpath, query_params
            )
        )
        json_body = {
            "DATA_TYPE": "rename",
            "old_path": oldpath,
            "new_path": newpath,
            "local_user": local_user,
        }
        return self.post(
            f"/v0.10/operation/endpoint/{endpoint_id}/rename",
            data=json_body,
            query_params=query_params,
        )

    def operation_stat(
        self,
        endpoint_id: uuid.UUID | str,
        path: str | MissingType = MISSING,
        *,
        local_user: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param endpoint_id: The ID of the collection on which to do the stat operation
        :param path: Path on the collection to do the stat operation on
        :param local_user: Optional value passed to identity mapping specifying which
            local user account to map to. Only usable with Globus Connect Server v5
            mapped collections.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    tc.operation_stat(ep_id, "/path/to/item")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: transfer.operation_stat

            .. tab-item:: API Info

                ``GET /operation/endpoint/<endpoint_id>/stat``

                .. extdoclink:: Get File or Directory Status
                    :ref: transfer/file_operations/#stat
        """
        query_params = {
            "path": path,
            "local_user": local_user,
            **(query_params or {}),
        }
        log.debug(f"TransferClient.operation_stat({endpoint_id}, {query_params})")
        return self.get(
            f"/v0.10/operation/endpoint/{endpoint_id}/stat", query_params=query_params
        )

    #
    # Task Submission
    #

    def get_submission_id(
        self, *, query_params: dict[str, t.Any] | None = None
    ) -> response.GlobusHTTPResponse:
        """
        :param query_params: Additional passthrough query parameters

        Submission IDs are required to submit tasks to the Transfer service
        via the :meth:`submit_transfer <.submit_transfer>` and
        :meth:`submit_delete <.submit_delete>` methods.

        Most users will not need to call this method directly, as the
        methods :meth:`~submit_transfer` and :meth:`~submit_delete` will call it
        automatically if the data does not contain a ``submission_id``.

        .. tab-set::

            .. tab-item:: API Info

                ``GET /submission_id``

                .. extdoclink:: Get a submission ID
                    :ref: transfer/task_submit/#get_submission_id

            .. tab-item:: Example Response Data

                .. expandtestfixture:: transfer.get_submission_id
        """
        log.debug(f"TransferClient.get_submission_id({query_params})")
        return self.get("/v0.10/submission_id", query_params=query_params)

    def submit_transfer(
        self, data: dict[str, t.Any] | TransferData
    ) -> response.GlobusHTTPResponse:
        """
        :param data: A transfer task document listing files and directories, and setting
            various options. See :class:`TransferData <globus_sdk.TransferData>` for
            details

        Submit a Transfer Task.

        If no ``submission_id`` is included in the payload, one will be requested and
        used automatically. The data passed to this method will be modified to include
        the ``submission_id``.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tdata = globus_sdk.TransferData(
                        source_endpoint_id,
                        destination_endpoint_id,
                        label="SDK example",
                        sync_level="checksum",
                    )
                    tdata.add_item("/source/path/dir/", "/dest/path/dir/")
                    tdata.add_item("/source/path/file.txt", "/dest/path/file.txt")

                    tc = globus_sdk.TransferClient(...)
                    transfer_result = tc.submit_transfer(tdata)
                    print("task_id =", transfer_result["task_id"])

                The `data` parameter can be a normal Python dictionary, or
                a :class:`TransferData <globus_sdk.TransferData>` object.

            .. tab-item:: API Info

                ``POST /transfer``

                .. extdoclink:: Submit a transfer task
                    :ref: transfer/task_submit/#submit_transfer_task
        """  # noqa: E501
        log.debug("TransferClient.submit_transfer(...)")
        if "submission_id" not in data or data["submission_id"] is MISSING:
            log.debug("submit_transfer autofetching submission_id")
            data["submission_id"] = self.get_submission_id()["value"]
        return self.post("/v0.10/transfer", data=data)

    def submit_delete(
        self, data: dict[str, t.Any] | DeleteData
    ) -> response.GlobusHTTPResponse:
        """
        :param data: A delete task document listing files and directories, and setting
            various options. See :class:`DeleteData <globus_sdk.DeleteData>` for
            details

        Submit a Delete Task.

        If no ``submission_id`` is included in the payload, one will be requested and
        used automatically. The data passed to this method will be modified to include
        the ``submission_id``.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    ddata = globus_sdk.DeleteData(endpoint_id, recursive=True)
                    ddata.add_item("/dir/to/delete/")
                    ddata.add_item("/file/to/delete/file.txt")

                    tc = globus_sdk.TransferClient(...)
                    delete_result = tc.submit_delete(ddata)
                    print("task_id =", delete_result["task_id"])

                The `data` parameter can be a normal Python dictionary, or
                a :class:`DeleteData <globus_sdk.DeleteData>` object.

            .. tab-item:: API Info

                ``POST /delete``

                .. extdoclink:: Submit a delete task
                    :ref: transfer/task_submit/#submit_delete_task
        """
        log.debug("TransferClient.submit_delete(...)")
        if "submission_id" not in data or data["submission_id"] is MISSING:
            log.debug("submit_delete autofetching submission_id")
            data["submission_id"] = self.get_submission_id()["value"]
        return self.post("/v0.10/delete", data=data)

    #
    # Task inspection and management
    #

    @paging.has_paginator(
        paging.LimitOffsetTotalPaginator,
        items_key="DATA",
        get_page_size=_get_page_size,
        max_total_results=1000,
        page_size=1000,
    )
    def task_list(
        self,
        *,
        limit: int | MissingType = MISSING,
        offset: int | MissingType = MISSING,
        orderby: str | list[str] | MissingType = MISSING,
        # pylint: disable=redefined-builtin
        filter: str | TransferFilterDict | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        Get an iterable of task documents owned by the current user.

        :param limit: limit the number of results
        :param offset: offset used in paging
        :param orderby: One or more order-by options. Each option is
            either a field name or a field name followed by a space and 'ASC' or 'DESC'
            for ascending or descending. See example usage for details.
        :param filter: Only return task documents which match these filter clauses. For
            the filter syntax, see the **External Documentation** linked below. If a
            dict is supplied as the filter, it is formatted as a set of filter clauses.
            See :ref:`filter formatting <transfer_filter_formatting>` for details.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                Fetch 10 tasks and print some basic info:

                .. code-block:: python

                    tc = TransferClient(...)
                    for task in tc.task_list(limit=10):
                        print(
                            f"Task({task['task_id']}): "
                            f"{task['source_endpoint']} -> {task['destination_endpoint']}"
                        )

                Fetch 3 *specific* tasks using a ``task_id`` filter:

                .. code-block:: python

                    tc = TransferClient(...)
                    task_ids = [
                        "acb4b581-b3f3-403a-a42a-9da97aaa9961",
                        "39447a3c-e002-401a-b95c-f48b69b4c60a",
                        "02330d3a-987b-4abb-97ed-6a22f8fa365e",
                    ]
                    for task in tc.task_list(filter={"task_id": task_ids}):
                        print(
                            f"Task({task['task_id']}): "
                            f"{task['source_endpoint']} -> {task['destination_endpoint']}"
                        )

                Fetch the last 5 failed tasks by submission time:

                ..  code-block:: python

                    tc = TransferClient(...)
                    for task in tc.task_list(
                        limit=5,
                        orderby="request_time DESC",
                        filter={"status": "FAILED"}
                    ):
                        print(
                            f"Task({task['task_id']}) failed: "
                            f"{task['source_endpoint']} -> {task['destination_endpoint']}"
                        )


            .. tab-item:: Paginated Usage

                .. paginatedusage:: task_list

            .. tab-item:: API Info

                ``GET /task_list``

                .. extdoclink:: Task List
                    :ref: transfer/task/#get_task_list
        """  # noqa: E501
        log.debug("TransferClient.task_list(...)")
        query_params = {
            "limit": limit,
            "offset": offset,
            "orderby": commajoin(orderby),
            "filter": _format_filter_item(filter),
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get("/v0.10/task_list", query_params=query_params)
        )

    @paging.has_paginator(
        paging.LimitOffsetTotalPaginator,
        items_key="DATA",
        get_page_size=_get_page_size,
        max_total_results=1000,
        page_size=1000,
    )
    def task_event_list(
        self,
        task_id: uuid.UUID | str,
        *,
        limit: int | MissingType = MISSING,
        offset: int | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        r"""
        List events (for example, faults and errors) for a given Task.

        :param task_id: The ID of the task to inspect
        :param limit: limit the number of results
        :param offset: offset used in paging
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                Fetch 10 events and print some basic info:

                .. code-block:: python

                    tc = TransferClient(...)
                    task_id = ...
                    for event in tc.task_event_list(task_id, limit=10):
                        print(f"Event on Task({task_id}) at {event['time']}:\n{event['description']}")

            .. tab-item:: Paginated Usage

                .. paginatedusage:: task_event_list

            .. tab-item:: API Info

                ``GET /task/<task_id>/event_list``

                .. extdoclink:: Get Event List
                    :ref: transfer/task/#get_event_list
        """  # noqa: E501
        log.debug(f"TransferClient.task_event_list({task_id}, ...)")
        query_params = {
            "limit": limit,
            "offset": offset,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(f"/v0.10/task/{task_id}/event_list", query_params=query_params)
        )

    def get_task(
        self,
        task_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        :param task_id: The ID of the task to inspect
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /task/<task_id>``

                .. extdoclink:: Get task by ID
                    :ref: transfer/task/#get_task_by_id
        """
        log.debug(f"TransferClient.get_task({task_id}, ...)")
        return self.get(f"/v0.10/task/{task_id}", query_params=query_params)

    def update_task(
        self,
        task_id: uuid.UUID | str,
        data: dict[str, t.Any],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Modify a task. Only tasks which are still running can be modified, and only the
        ``label`` and ``deadline`` fields can be updated.

        :param task_id: The ID of the task to modify
        :param data: A partial task document with fields to update
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PUT /task/<task_id>``

                .. extdoclink:: Update task by ID
                    :ref: transfer/task/#update_task_by_id
        """
        log.debug(f"TransferClient.update_task({task_id}, ...)")
        return self.put(f"/v0.10/task/{task_id}", data=data, query_params=query_params)

    def cancel_task(self, task_id: uuid.UUID | str) -> response.GlobusHTTPResponse:
        """
        Cancel a task which is still running.

        :param task_id: The ID of the task to cancel

        .. tab-set::

            .. tab-item:: API Info

                ``POST /task/<task_id>/cancel``

                .. extdoclink:: Cancel task by ID
                    :ref: transfer/task/#cancel_task_by_id
        """
        log.debug(f"TransferClient.cancel_task({task_id})")
        return self.post(f"/v0.10/task/{task_id}/cancel")

    def task_wait(
        self, task_id: uuid.UUID | str, *, timeout: int = 10, polling_interval: int = 10
    ) -> bool:
        r"""
        Wait until a Task is complete or fails, with a time limit. If the task
        is "ACTIVE" after time runs out, returns ``False``. Otherwise returns
        ``True``.

        :param task_id: ID of the Task to wait on for completion
        :param timeout: Number of seconds to wait in total. Minimum 1. [Default: ``10``]
        :param polling_interval: Number of seconds between queries to Globus about the
            Task status. Minimum 1. [Default: ``10``]

        .. tab-set::

            .. tab-item:: Example Usage

                If you want to wait for a task to terminate, but want to warn every
                minute that it doesn't terminate, you could:

                .. code-block:: python

                    tc = TransferClient(...)
                    while not tc.task_wait(task_id, timeout=60):
                        print(f"Another minute went by without {task_id} terminating")

                Or perhaps you want to check on a task every minute for 10 minutes, and
                give up if it doesn't complete in that time:

                .. code-block:: python

                    tc = TransferClient(...)
                    done = tc.task_wait(task_id, timeout=600, polling_interval=60)
                    if not done:
                        print(f"{task_id} didn't successfully terminate!")
                    else:
                        print(f"{task_id} completed")

                You could print dots while you wait for a task by only waiting one
                second at a time:

                .. code-block:: python

                    tc = TransferClient(...)
                    while not tc.task_wait(task_id, timeout=1, polling_interval=1):
                        print(".", end="")
                    print(f"\n{task_id} completed!")
        """  # noqa: E501
        log.debug(
            "TransferClient.task_wait(%s, %s, %s)", task_id, timeout, polling_interval
        )

        # check valid args
        if timeout < 1:
            log.error(f"task_wait() timeout={timeout} is less than minimum of 1s")
            raise exc.GlobusSDKUsageError(
                "TransferClient.task_wait timeout has a minimum of 1"
            )
        if polling_interval < 1:
            log.error(
                "task_wait() polling_interval={} is less than minimum of 1s".format(
                    polling_interval
                )
            )
            raise exc.GlobusSDKUsageError(
                "TransferClient.task_wait polling_interval has a minimum of 1"
            )

        # ensure that we always wait at least one interval, even if the timeout
        # is shorter than the polling interval, by reducing the interval to the
        # timeout if it is larger
        polling_interval = min(timeout, polling_interval)

        # helper for readability
        def timed_out(waited_time: int) -> bool:
            return waited_time > timeout

        waited_time = 0
        # doing this as a while-True loop actually makes it simpler than doing
        # while not timed_out(waited_time) because of the end condition
        while True:
            # get task, check if status != ACTIVE
            task = self.get_task(task_id)
            status = task["status"]
            if status != "ACTIVE":
                log.debug(
                    "task_wait(task_id={}) terminated with status={}".format(
                        task_id, status
                    )
                )
                return True

            # make sure to check if we timed out before sleeping again, so we
            # don't sleep an extra polling_interval
            waited_time += polling_interval
            if timed_out(waited_time):
                log.debug(f"task_wait(task_id={task_id}) timed out")
                return False

            log.debug(f"task_wait(task_id={task_id}) waiting {polling_interval}s")
            time.sleep(polling_interval)
        # unreachable -- end of task_wait

    def task_pause_info(
        self,
        task_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get info about why a task is paused or about to be paused.

        :param task_id: The ID of the task to inspect
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /task/<task_id>/pause_info``

                .. extdoclink:: Get task pause info
                    :ref: transfer/task/#get_task_pause_info
        """
        log.debug(f"TransferClient.task_pause_info({task_id}, ...)")
        return self.get(f"/v0.10/task/{task_id}/pause_info", query_params=query_params)

    @paging.has_paginator(
        paging.NullableMarkerPaginator, items_key="DATA", marker_key="next_marker"
    )
    def task_successful_transfers(
        self,
        task_id: uuid.UUID | str,
        *,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        Get the successful file transfers for a completed Task.

        .. note::

            Only files that were actually transferred are included. This does
            not include directories, files that were checked but skipped as
            part of a sync transfer, or files which were skipped due to
            skip_source_errors being set on the task.

        :param task_id: The ID of the task to inspect
        :param marker: A marker for pagination
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                Fetch all transferred files for a task and print some basic info:

                .. code-block:: python

                    tc = TransferClient(...)
                    task_id = ...
                    for info in tc.task_successful_transfers(task_id):
                        print(f"{info['source_path']} -> {info['destination_path']}")

            .. tab-item:: Paginated Usage

                .. paginatedusage:: task_successful_transfers

            .. tab-item:: API Info

                ``GET /task/<task_id>/successful_transfers``

                .. extdoclink:: Get Task Successful Transfers
                    :ref: transfer/task/#get_task_successful_transfers
        """  # noqa: E501
        log.debug(f"TransferClient.task_successful_transfers({task_id}, ...)")
        query_params = {
            "marker": marker,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(
                f"/v0.10/task/{task_id}/successful_transfers", query_params=query_params
            )
        )

    @paging.has_paginator(
        paging.NullableMarkerPaginator, items_key="DATA", marker_key="next_marker"
    )
    def task_skipped_errors(
        self,
        task_id: uuid.UUID | str,
        *,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        Get path and error information for all paths that were skipped due
        to skip_source_errors being set on a completed transfer Task.

        :param task_id: The ID of the task to inspect
        :param marker: A marker for pagination
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                Fetch all skipped errors for a task and print some basic info:

                .. code-block:: python

                    tc = TransferClient(...)
                    task_id = ...
                    for info in tc.task_skipped_errors(task_id):
                        print(f"{info['error_code']} -> {info['source_path']}")

            .. tab-item:: Paginated Usage

                .. paginatedusage:: task_skipped_errors

            .. tab-item:: API Info

                ``GET /task/<task_id>/skipped_errors``

                .. extdoclink:: Get Task Skipped Errors
                    :ref: transfer/task/#get_task_skipped_errors
        """  # noqa: E501
        log.debug("TransferClient.task_skipped_errors(%s, ...)", task_id)
        query_params = {
            "marker": marker,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(f"/v0.10/task/{task_id}/skipped_errors", query_params=query_params)
        )

    #
    # advanced endpoint management (requires endpoint manager role)
    #

    def endpoint_manager_monitored_endpoints(
        self, *, query_params: dict[str, t.Any] | None = None
    ) -> IterableTransferResponse:
        """
        Get endpoints the current user is a monitor or manager on.

        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET endpoint_manager/monitored_endpoints``

                .. extdoclink:: Get monitored endpoints and collections
                    :ref: transfer/advanced_collection_management/#get_monitored_endpoints_and_collections
        """  # noqa: E501
        log.debug(
            f"TransferClient.endpoint_manager_monitored_endpoints({query_params})"
        )
        return IterableTransferResponse(
            self.get(
                "/v0.10/endpoint_manager/monitored_endpoints", query_params=query_params
            )
        )

    def endpoint_manager_hosted_endpoint_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        Get shared endpoints hosted on the given endpoint.

        :param endpoint_id: The ID of the host endpoint
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint_manager/endpoint/<endpoint_id>/hosted_endpoint_list``

                .. extdoclink:: Get hosted endpoint list
                    :ref: transfer/advanced_collection_management/#get_hosted_collection_list
        """  # noqa: E501
        log.debug(
            f"TransferClient.endpoint_manager_hosted_endpoint_list({endpoint_id})"
        )
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint_manager/endpoint/{endpoint_id}/hosted_endpoint_list",
                query_params=query_params,
            )
        )

    def endpoint_manager_get_endpoint(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get endpoint details as an admin.

        :param endpoint_id: The ID of the endpoint
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint_manager/endpoint/<endpoint_id>``

                .. extdoclink:: Get endpoint as admin
                    :ref: transfer/advanced_collection_management/#mc_get_endpoint_or_collection
        """  # noqa: E501
        log.debug(f"TransferClient.endpoint_manager_get_endpoint({endpoint_id})")
        return self.get(
            f"/v0.10/endpoint_manager/endpoint/{endpoint_id}", query_params=query_params
        )

    def endpoint_manager_acl_list(
        self,
        endpoint_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        Get a list of access control rules on specified endpoint as an admin.

        :param endpoint_id: The ID of the endpoint
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET endpoint_manager/endpoint/<endpoint_id>/access_list``

                .. extdoclink:: Get guest collection access list as admin
                    :ref: transfer/advanced_collection_management/#get_guest_collection_access_list_as_admin
        """  # noqa: E501
        log.debug(
            f"TransferClient.endpoint_manager_endpoint_acl_list({endpoint_id}, ...)"
        )
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint_manager/endpoint/{endpoint_id}/access_list",
                query_params=query_params,
            )
        )

    #
    # endpoint manager task methods
    #

    @paging.has_paginator(paging.LastKeyPaginator, items_key="DATA")
    def endpoint_manager_task_list(
        self,
        *,
        filter_status: str | t.Iterable[str] | MissingType = MISSING,
        filter_task_id: (
            uuid.UUID | str | t.Iterable[uuid.UUID | str] | MissingType
        ) = MISSING,
        filter_owner_id: uuid.UUID | str | MissingType = MISSING,
        filter_endpoint: uuid.UUID | str | MissingType = MISSING,
        filter_endpoint_use: t.Literal["source", "destination"] | MissingType = MISSING,
        filter_is_paused: bool | MissingType = MISSING,
        filter_completion_time: str | tuple[DateLike, DateLike] | MissingType = MISSING,
        filter_min_faults: int | MissingType = MISSING,
        filter_local_user: str | MissingType = MISSING,
        last_key: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        r"""
        Get a list of tasks visible via ``activity_monitor`` role, as opposed
        to tasks owned by the current user.

        For any query that doesn't specify a ``filter_status`` that is a subset of
        ``("ACTIVE", "INACTIVE")``, at least one of ``filter_task_id`` or
        ``filter_endpoint`` is required.

        :param filter_status: Return only tasks with any of the specified statuses
            Note that in-progress tasks will have status ``"ACTIVE"`` or ``"INACTIVE"``,
            and completed tasks will have status ``"SUCCEEDED"`` or ``"FAILED"``.
        :param filter_task_id: Return only tasks with any of the specified ids. If any
            of the specified tasks do not involve an endpoint the user has an
            appropriate role for, a ``PermissionDenied`` error will be returned. This
            filter can't be combined with any other filter.  If another filter is
            passed, a ``BadRequest`` will be returned. (limit: 50 task IDs)
        :param filter_owner_id: A Globus Auth identity id. Limit results to tasks
            submitted by the specified identity, or linked to the specified identity,
            at submit time.  Returns ``UserNotFound`` if the identity does not exist or
            has never used the Globus Transfer service. If no tasks were submitted by
            this user to an endpoint the current user has an appropriate role on, an
            empty result set will be returned. Unless filtering for running tasks (i.e.
            ``filter_status`` is a subset of ``("ACTIVE", "INACTIVE")``,
            ``filter_endpoint`` is required when using ``filter_owner_id``.
        :param filter_endpoint: Single endpoint id. Return only tasks with a matching
            source or destination endpoint or matching source or destination host
            endpoint.
        :param filter_endpoint_use: In combination with ``filter_endpoint``, filter to
            tasks where the endpoint or collection was used specifically as the source or
            destination of the transfer.
        :param filter_is_paused: Return only tasks with the specified ``is_paused``
            value. Requires that ``filter_status`` is also passed and contains a subset
            of ``"ACTIVE"`` and ``"INACTIVE"``. Completed tasks always have
            ``is_paused`` equal to ``False`` and filtering on their paused state is not
            useful and not supported.  Note that pausing is an async operation, and
            after a pause rule is inserted it will take time before the ``is_paused``
            flag is set on all affected tasks. Tasks paused by id will have the
            ``is_paused`` flag set immediately.
        :param filter_completion_time: Start and end date-times separated by a comma, or
            provided as a tuple of strings or datetime objects.  Returns only completed
            tasks with ``completion_time`` in the specified range. Date strings should
            be specified in one of the following ISO 8601 formats:
            ``YYYY-MM-DDTHH:MM:SS``, ``YYYY-MM-DDTHH:MM:SS+/-HH:MM``, or
            ``YYYY-MM-DDTHH:MM:SSZ``. If no timezone is specified, UTC is assumed. A
            space can be used between the date and time instead of ``T``. A blank string
            may be used for either the start or end (but not both) to indicate no limit
            on that side. If the end date is blank, the filter will also include all
            active tasks, since they will complete some time in the future.
        :param filter_min_faults:  Minimum number of cumulative faults, inclusive.
            Return only tasks with ``faults >= N``, where ``N`` is the filter value.
            Use ``filter_min_faults=1`` to find all tasks with at least one fault.
            Note that many errors are not fatal and the task may still be successful
            even if ``faults >= 1``.
        :param filter_local_user: A valid username for the target system running the
            endpoint, as a utf8 encoded string. Requires that ``filter_endpoint`` is
            also set. Return only tasks that have successfully fetched the local user
            from the endpoint, and match the values of ``filter_endpoint`` and
            ``filter_local_user`` on the source or on the destination.
        :param last_key: the last key, for paging
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Example Usage

                Fetch some tasks and print some basic info:

                .. code-block:: python

                    tc = TransferClient(...)
                    for task in tc.endpoint_manager_task_list(filter_status="ACTIVE"):
                        print(
                            f"Task({task['task_id']}): "
                            f"{task['source_endpoint']} -> {task['destination_endpoint']}\n"
                            "  was submitted by\n"
                            f"  {task['owner_string']}"
                        )

            .. tab-item:: Paginated Usage

                .. paginatedusage:: endpoint_manager_task_list

                For example, fetch and print all active tasks visible via
                ``activity_monitor`` permissions:

                .. code-block:: python

                    tc = TransferClient(...)
                    for page in tc.paginated.endpoint_manager_task_list(filter_status="ACTIVE"):
                        for task in page:
                            print(
                                f"Task({task['task_id']}): "
                                f"{task['source_endpoint']} -> {task['destination_endpoint']}\n"
                                "  was submitted by\n"
                                f"  {task['owner_string']}"
                            )

            .. tab-item:: API Info

                ``GET endpoint_manager/task_list``

                .. extdoclink:: Advanced Collection Management: Get tasks
                    :ref: transfer/advanced_collection_management/#get_tasks
        """  # noqa: E501
        log.debug("TransferClient.endpoint_manager_task_list(...)")
        if filter_endpoint is MISSING and filter_endpoint_use is not MISSING:
            raise exc.GlobusSDKUsageError(
                "`filter_endpoint_use` is only valid when `filter_endpoint` is "
                "also supplied."
            )
        query_params = {
            "filter_status": commajoin(filter_status),
            "filter_task_id": commajoin(filter_task_id),
            "filter_owner_id": filter_owner_id,
            "filter_endpoint": filter_endpoint,
            "filter_endpoint_use": filter_endpoint_use,
            "filter_is_paused": filter_is_paused,
            "filter_completion_time": _format_completion_time(filter_completion_time),
            "filter_min_faults": filter_min_faults,
            "filter_local_user": filter_local_user,
            "last_key": last_key,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get("/v0.10/endpoint_manager/task_list", query_params=query_params)
        )

    def endpoint_manager_get_task(
        self,
        task_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get task info as an admin. Requires activity monitor effective role on
        the destination endpoint of the task.

        :param task_id: The ID of the task to inspect
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint_manager/task/<task_id>``

                .. extdoclink:: Advanced Collection Management: Get task
                    :ref: transfer/advanced_collection_management/#get_task
        """
        log.debug(f"TransferClient.endpoint_manager_get_task({task_id}, ...)")
        return self.get(
            f"/v0.10/endpoint_manager/task/{task_id}", query_params=query_params
        )

    @paging.has_paginator(
        paging.LimitOffsetTotalPaginator,
        items_key="DATA",
        get_page_size=_get_page_size,
        max_total_results=1000,
        page_size=1000,
    )
    def endpoint_manager_task_event_list(
        self,
        task_id: uuid.UUID | str,
        *,
        limit: int | MissingType = MISSING,
        offset: int | MissingType = MISSING,
        filter_is_error: bool | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        List events (for example, faults and errors) for a given task as an
        admin. Requires activity monitor effective role on the destination
        endpoint of the task.

        :param task_id: The ID of the task to inspect
        :param limit: limit the number of results
        :param offset: offset used in paging
        :param filter_is_error: Return only events that are errors. A value of ``False``
            (returning only non-errors) is not supported. By default all events are
            returned.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Paginated Usage

                .. paginatedusage:: endpoint_manager_task_event_list

            .. tab-item:: API Info

                ``GET /task/<task_id>/event_list``

                .. extdoclink:: Advanced Collection Management: Get task events
                    :ref: transfer/advanced_collection_management/#get_task_events
        """
        log.debug(f"TransferClient.endpoint_manager_task_event_list({task_id}, ...)")
        query_params = {
            "limit": limit,
            "offset": offset,
            "filter_is_error": (
                1
                if filter_is_error
                else 0 if isinstance(filter_is_error, bool) else filter_is_error
            ),
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint_manager/task/{task_id}/event_list",
                query_params=query_params,
            )
        )

    def endpoint_manager_task_pause_info(
        self,
        task_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get details about why a task is paused as an admin. Requires activity
        monitor effective role on the destination endpoint of the task.

        :param task_id: The ID of the task to inspect
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint_manager/task/<task_id>/pause_info``

                .. extdoclink:: Get task pause info as admin
                    :ref: transfer/advanced_collection_management/#get_task_pause_info_as_admin
        """  # noqa: E501
        log.debug(f"TransferClient.endpoint_manager_task_pause_info({task_id}, ...)")
        return self.get(
            f"/v0.10/endpoint_manager/task/{task_id}/pause_info",
            query_params=query_params,
        )

    @paging.has_paginator(
        paging.NullableMarkerPaginator, items_key="DATA", marker_key="next_marker"
    )
    def endpoint_manager_task_successful_transfers(
        self,
        task_id: uuid.UUID | str,
        *,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        r"""
        Get the successful file transfers for a completed Task as an admin.

        :param task_id: The ID of the task to inspect
        :param marker: A marker for pagination
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Paginated Usage

                .. paginatedusage:: endpoint_manager_task_successful_transfers

            .. tab-item:: API Info

                ``GET /endpoint_manager/task/<task_id>/successful_transfers``

                .. extdoclink:: Get task successful transfers as admin
                    :ref: transfer/advanced_collection_management/\
                            #get_task_successful_transfers_as_admin
        """
        log.debug(
            "TransferClient.endpoint_manager_task_successful_transfers(%s, ...)",
            task_id,
        )
        query_params = {
            "marker": marker,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint_manager/task/{task_id}/successful_transfers",
                query_params=query_params,
            )
        )

    @paging.has_paginator(
        paging.NullableMarkerPaginator, items_key="DATA", marker_key="next_marker"
    )
    def endpoint_manager_task_skipped_errors(
        self,
        task_id: uuid.UUID | str,
        *,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        r"""
        Get skipped errors for a completed Task as an admin.

        :param task_id: The ID of the task to inspect
        :param marker: A marker for pagination
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Paginated Usage

                .. paginatedusage:: endpoint_manager_task_skipped_errors

            .. tab-item:: API Info

                ``GET /endpoint_manager/task/<task_id>/skipped_errors``

                .. extdoclink:: Get task skipped errors as admin
                    :ref: transfer/advanced_collection_management/\
                            #get_task_skipped_errors_as_admin
        """
        log.debug(
            f"TransferClient.endpoint_manager_task_skipped_errors({task_id}, ...)"
        )
        query_params = {
            "marker": marker,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(
                f"/v0.10/endpoint_manager/task/{task_id}/skipped_errors",
                query_params=query_params,
            )
        )

    def endpoint_manager_cancel_tasks(
        self,
        task_ids: t.Iterable[uuid.UUID | str],
        message: str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Cancel a list of tasks as an admin. Requires activity manager effective
        role on the task(s) source or destination endpoint(s).

        :param task_ids: List of task ids to cancel
        :param message: Message given to all users whose tasks have been canceled
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``POST /endpoint_manager/admin_cancel``

                .. extdoclink:: Cancel tasks as admin
                    :ref: transfer/advanced_collection_management/#admin_cancel
        """
        str_task_ids = [str(i) for i in task_ids]
        log.debug(
            f"TransferClient.endpoint_manager_cancel_tasks({str_task_ids}, {message})"
        )
        data = {"message": message, "task_id_list": str_task_ids}
        return self.post(
            "/v0.10/endpoint_manager/admin_cancel", data=data, query_params=query_params
        )

    def endpoint_manager_cancel_status(
        self,
        admin_cancel_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get the status of an an admin cancel (result of
        endpoint_manager_cancel_tasks).

        :param admin_cancel_id: The ID of the the cancel job to inspect
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint_manager/admin_cancel/<admin_cancel_id>``

                .. extdoclink:: Get cancel status by ID
                    :ref: transfer/advanced_collection_management/#get_cancel_status_by_id
        """  # noqa: E501
        log.debug(f"TransferClient.endpoint_manager_cancel_status({admin_cancel_id})")
        return self.get(
            f"/v0.10/endpoint_manager/admin_cancel/{admin_cancel_id}",
            query_params=query_params,
        )

    def endpoint_manager_pause_tasks(
        self,
        task_ids: t.Iterable[uuid.UUID | str],
        message: str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Pause a list of tasks as an admin. Requires activity manager effective
        role on the task(s) source or destination endpoint(s).

        :param task_ids: List of task ids to pause
        :param message: Message given to all users whose tasks have been paused
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``POST /endpoint_manager/admin_pause``

                .. extdoclink:: Pause tasks as admin
                    :ref: transfer/advanced_collection_management/#pause_tasks_as_admin
        """
        str_task_ids = [str(i) for i in task_ids]
        log.debug(
            f"TransferClient.endpoint_manager_pause_tasks({str_task_ids}, {message})"
        )
        data = {"message": message, "task_id_list": str_task_ids}
        return self.post(
            "/v0.10/endpoint_manager/admin_pause", data=data, query_params=query_params
        )

    def endpoint_manager_resume_tasks(
        self,
        task_ids: t.Iterable[uuid.UUID | str],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Resume a list of tasks as an admin. Requires activity manager effective
        role on the task(s) source or destination endpoint(s).

        :param task_ids: List of task ids to resume
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``POST /endpoint_manager/admin_resume``

                .. extdoclink:: Resume tasks as admin
                    :ref: transfer/advanced_collection_management/#resume_tasks_as_admin
        """
        str_task_ids = [str(i) for i in task_ids]
        log.debug(f"TransferClient.endpoint_manager_resume_tasks({str_task_ids})")
        data = {"task_id_list": str_task_ids}
        return self.post(
            "/v0.10/endpoint_manager/admin_resume", data=data, query_params=query_params
        )

    #
    # endpoint manager pause rule methods
    #

    def endpoint_manager_pause_rule_list(
        self,
        *,
        filter_endpoint: uuid.UUID | str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableTransferResponse:
        """
        Get a list of pause rules on endpoints that the current user has the
        activity monitor effective role on.

        :param filter_endpoint: An endpoint ID. Limit results to rules on endpoints
            hosted by this endpoint. Must be activity monitor on this endpoint, not just
            the hosted endpoints.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint_manager/pause_rule_list``

                .. extdoclink:: Get pause rules
                    :ref: transfer/advanced_collection_management/#get_pause_rules
        """
        log.debug("TransferClient.endpoint_manager_pause_rule_list(...)")
        query_params = {
            "filter_endpoint": filter_endpoint,
            **(query_params or {}),
        }
        return IterableTransferResponse(
            self.get(
                "/v0.10/endpoint_manager/pause_rule_list", query_params=query_params
            )
        )

    def endpoint_manager_create_pause_rule(
        self, data: dict[str, t.Any] | None
    ) -> response.GlobusHTTPResponse:
        """
        Create a new pause rule. Requires the activity manager effective role
        on the endpoint defined in the rule.

        :param data: A pause rule document describing the rule to create

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    rule_data = {
                        "DATA_TYPE": "pause_rule",
                        "message": "Message to users explaining why tasks are paused",
                        "endpoint_id": "339abc22-aab3-4b45-bb56-8d40535bfd80",
                        "identity_id": None,  # affect all users on endpoint
                        "start_time": None,  # start now
                    }
                    create_result = tc.endpoint_manager_create_pause_rule(ep_data)
                    rule_id = create_result["id"]

            .. tab-item:: API Info

                ``POST /endpoint_manager/pause_rule``

                .. extdoclink:: Create pause rule
                    :ref: transfer/advanced_collection_management/#create_pause_rule
        """
        log.debug("TransferClient.endpoint_manager_create_pause_rule(...)")
        return self.post("/v0.10/endpoint_manager/pause_rule", data=data)

    def endpoint_manager_get_pause_rule(
        self,
        pause_rule_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get an existing pause rule by ID. Requires the activity manager
        effective role on the endpoint defined in the rule.

        :param pause_rule_id: ID of pause rule to get
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint_manager/pause_rule/<pause_rule_id>``

                .. extdoclink:: Get pause rule
                    :ref: transfer/advanced_collection_management/#get_pause_rule
        """
        log.debug(f"TransferClient.endpoint_manager_get_pause_rule({pause_rule_id})")
        return self.get(
            f"/v0.10/endpoint_manager/pause_rule/{pause_rule_id}",
            query_params=query_params,
        )

    def endpoint_manager_update_pause_rule(
        self,
        pause_rule_id: uuid.UUID | str,
        data: dict[str, t.Any] | None,
    ) -> response.GlobusHTTPResponse:
        """
        Update an existing pause rule by ID. Requires the activity manager
        effective role on the endpoint defined in the rule.
        Note that non update-able fields in data will be ignored.

        :param pause_rule_id: The ID of the pause rule to update
        :param data: A partial pause rule document with fields to update

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    rule_data = {
                        "message": "Update to pause, reads are now allowed.",
                        "pause_ls": False,
                        "pause_task_transfer_read": False,
                    }
                    update_result = tc.endpoint_manager_update_pause_rule(ep_data)

            .. tab-item:: API Info

                ``PUT /endpoint_manager/pause_rule/<pause_rule_id>``

                .. extdoclink:: Update pause rule
                    :ref: transfer/advanced_collection_management/#update_pause_rule
        """
        log.debug(f"TransferClient.endpoint_manager_update_pause_rule({pause_rule_id})")
        return self.put(
            f"/v0.10/endpoint_manager/pause_rule/{pause_rule_id}", data=data
        )

    def endpoint_manager_delete_pause_rule(
        self,
        pause_rule_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete an existing pause rule by ID. Requires the user to see the
        "editable" field of the rule as True. Any tasks affected by this rule
        will no longer be once it is deleted.

        :param pause_rule_id: The ID of the pause rule to delete
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /endpoint_manager/pause_rule/<pause_rule_id>``

                .. extdoclink:: Delete pause rule
                    :ref: transfer/advanced_collection_management/#delete_pause_rule
        """
        log.debug(f"TransferClient.endpoint_manager_delete_pause_rule({pause_rule_id})")
        return self.delete(
            f"/v0.10/endpoint_manager/pause_rule/{pause_rule_id}",
            query_params=query_params,
        )

    # Tunnel methods

    def create_tunnel(
        self,
        data: dict[str, t.Any] | CreateTunnelData,
    ) -> response.GlobusHTTPResponse:
        """
        .. note::

            Tunnels functionality is currently in Beta and may experience
            changes that will break this interface

        :param data: Parameters for the tunnel creation

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    result = tc.create_tunnel(data)
                    print(result["data"]["id"])

            .. tab-item:: API Info

                ``POST /v2/tunnels``
        """
        log.debug("TransferClient.create_tunnel(...)")
        try:
            data_element = data["data"]
        except KeyError as e:
            raise exc.GlobusSDKUsageError(
                "create_tunnel() body was malformed (missing the 'data' key). "
                "Use CreateTunnelData to easily create correct documents."
            ) from e

        try:
            attributes = data_element["attributes"]
        except KeyError:
            data_element["attributes"] = {}
            attributes = data_element["attributes"]
        if attributes.get("submission_id", MISSING) is MISSING:
            log.debug("create_tunnel auto-creating submission_id")
            attributes["submission_id"] = str(uuid.uuid1())

        r = self.post("/v2/tunnels", data=data)
        return r

    def update_tunnel(
        self,
        tunnel_id: str | uuid.UUID,
        update_doc: dict[str, t.Any],
    ) -> response.GlobusHTTPResponse:
        r"""
        .. note::

            Tunnels functionality is currently in Beta and may experience
            changes that will break this interface

        :param tunnel_id: The ID of the Tunnel.
        :param update_doc: The document that will be sent to the patch API.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    "data" = {
                        "type": "Tunnel",
                        "attributes": {
                            "state": "STOPPING",
                        },
                    }
                    result = tc.update_tunnel(tunnel_id, data)
                    print(result["data"])

            .. tab-item:: API Info

                ``PATCH /v2/tunnels/<tunnel_id>``
        """
        log.debug(f"TransferClient.update_tunnel({tunnel_id}, {update_doc})")
        r = self.patch(f"/v2/tunnels/{tunnel_id}", data=update_doc)
        return r

    def get_tunnel(
        self,
        tunnel_id: str | uuid.UUID,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        .. note::

            Tunnels functionality is currently in Beta and may experience
            changes that will break this interface

        :param tunnel_id: The ID of the Tunnel which we are fetching details about.
        :param query_params: Any additional parameters will be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    result = tc.show_tunnel(tunnel_id)
                    print(result["data"])

            .. tab-item:: API Info

                ``GET /v2/tunnels/<tunnel_id>``
        """
        log.debug("TransferClient.get_tunnel({tunnel_id}, {query_params})")
        r = self.get(f"/v2/tunnels/{tunnel_id}", query_params=query_params)
        return r

    def delete_tunnel(
        self,
        tunnel_id: str | uuid.UUID,
    ) -> response.GlobusHTTPResponse:
        """
        .. note::

            Tunnels functionality is currently in Beta and may experience
            changes that will break this interface

        :param tunnel_id: The ID of the Tunnel to be deleted.

        This will clean up all data associated with a Tunnel.
        Note that Tunnels must be stopped before they can be deleted.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    tc.delete_tunnel(tunnel_id)

            .. tab-item:: API Info

                ``DELETE /v2/tunnels/<tunnel_id>``
        """
        log.debug(f"TransferClient.delete_tunnel({tunnel_id})")
        r = self.delete(f"/v2/tunnels/{tunnel_id}")
        return r

    def list_tunnels(
        self,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableJSONAPIResponse:
        """
        .. note::

            Tunnels functionality is currently in Beta and may experience
            changes that will break this interface

        :param query_params: Any additional parameters will be passed through
            as query params.

        This will list all the Tunnels created by the authorized user.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    tc.list_tunnels(tunnel_id)

            .. tab-item:: API Info

                ``GET /v2/tunnels/``
        """
        log.debug(f"TransferClient.list_tunnels({query_params})")
        r = self.get("/v2/tunnels", query_params=query_params)
        return IterableJSONAPIResponse(r)

    def get_stream_access_point(
        self,
        stream_ap_id: str | uuid.UUID,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        .. note::

            Tunnels functionality is currently in Beta and may experience
            changes that will break this interface

        :param stream_ap_id: The ID of the steaming access point to lookup.
        :param query_params: Any additional parameters will be passed through
            as query params.

        Get a stream access point by id.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    tc.get_stream_ap(stream_ap_id)

            .. tab-item:: API Info

                ``GET /v2/stream_access_points/<stream_ap_id>``
        """
        log.debug(
            f"TransferClient.get_stream_access_point({stream_ap_id}, {query_params})"
        )
        r = self.get(
            f"/v2/stream_access_points/{stream_ap_id}", query_params=query_params
        )
        return r

    def get_tunnel_events(
        self,
        tunnel_id: str | uuid.UUID,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        .. note::

            Tunnels functionality is currently in Beta and may experience
            changes that will break this interface

        :param tunnel_id: The ID of the Tunnel which we are fetching events about.
        :param query_params: Any additional parameters will be passed through
            as query params.

        List all events for a specific tunnel.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    tc = globus_sdk.TransferClient(...)
                    result = tc.get_tunnel_events(tunnel_id)
                    print(result["data"])

            .. tab-item:: API Info

                ``GET /v2/tunnels/<tunnel_id>/events``
        """
        log.debug(f"TransferClient.get_tunnel_events({tunnel_id}, {query_params})")
        r = self.get(f"/v2/tunnels/{tunnel_id}/events", query_params=query_params)
        return r



import typing as t

from docutils.parsers.rst import directives

import globus_sdk

from ..utils import classname2methods, locate_class
from .add_content_directive import AddContentDirective


class EnumerateTestingFixtures(AddContentDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    option_spec = {
        "header_underline_char": directives.unchanged,
    }

    def gen_rst(self) -> t.Iterator[str]:
        from globus_sdk.testing import get_response_set

        underline_char = self.options.get("header_underline_char", "-")

        classname = self.arguments[0]
        cls = locate_class(classname)
        if not issubclass(cls, globus_sdk.BaseClient):
            msg = f"Expected {cls} to be a subclass of BaseClient"
            raise RuntimeError(msg)
        service_name = cls.service_name

        yield classname
        yield underline_char * len(classname)
        yield ""
        yield (
            f":class:`{classname}` has registered responses "
            "for the following methods:"
        )
        yield ""
        for methodname, method in classname2methods(classname, []):
            try:
                rset = get_response_set(method)
                # success -> has a response
            except ValueError:
                continue
                # error -> has no response, continue
            for casename in rset.cases():
                # use "attr" rather than "meth" so that sphinx does not add parens
                # the use of the method as an attribute of the class or instance better
                # matches how `testing` handles things
                yield (
                    ".. dropdown:: "
                    f':py:attr:`~{classname}.{methodname}` (``case="{casename}"``)'
                )
                yield ""
                yield f"    .. expandtestfixture:: {service_name}.{methodname}"
                yield f"       :case: {casename}"
                yield ""
        yield ""



from globus_sdk import GuestCollectionDocument, MappedCollectionDocument
from tests.common import register_api_route_fixture_file


def test_get_collection(client):
    register_api_route_fixture_file(
        "gcs", "/collections/COLLECTION_ID", "get_collection/normal.json"
    )
    res = client.get_collection("COLLECTION_ID")
    assert res["DATA_TYPE"] == "collection#1.0.0"
    assert res.full_data["DATA_TYPE"] == "result#1.0.0"
    assert "detail" in res.full_data
    assert "data" in res.full_data
    assert res.full_data["detail"] == "success"
    assert "detail" not in res.data
    assert res["display_name"] == "Happy Fun Collection Name"


def test_get_collection_flat(client):
    register_api_route_fixture_file(
        "gcs", "/collections/COLLECTION_ID", "get_collection/unexpectedly_flat.json"
    )
    res = client.get_collection("COLLECTION_ID")
    assert res["DATA_TYPE"] == "collection#1.0.0"
    assert res.full_data["DATA_TYPE"] == "collection#1.0.0"
    assert "detail" not in res.full_data
    assert "data" not in res.full_data
    assert res["display_name"] == "Happy Fun Collection Name"


def test_get_collection_bad_version(client):
    register_api_route_fixture_file(
        "gcs", "/collections/COLLECTION_ID", "get_collection/bad_version.json"
    )
    res = client.get_collection("COLLECTION_ID")
    assert res["DATA_TYPE"] == "result#1.0.0"
    assert res.full_data["DATA_TYPE"] == "result#1.0.0"
    assert "detail" in res.full_data
    assert "data" in res.full_data
    assert res.full_data["detail"] == "success"
    assert "detail" in res.data
    assert "foo" not in res.data
    for x in res.full_data["data"]:
        assert "foo" in x


def test_get_collection_includes_sideloaded_data(client):
    register_api_route_fixture_file(
        "gcs", "/collections/COLLECTION_ID", "get_collection/includes_other.json"
    )
    res = client.get_collection("COLLECTION_ID")
    assert res["DATA_TYPE"] == "collection#1.0.0"
    assert res.full_data["DATA_TYPE"] == "result#1.0.0"
    assert "detail" in res.full_data
    assert "data" in res.full_data
    assert res.full_data["detail"] == "success"
    assert "detail" not in res.data
    assert res["display_name"] == "Happy Fun Collection Name"


def test_get_collection_invalid_datatype_type(client):
    register_api_route_fixture_file(
        "gcs", "/collections/COLLECTION_ID", "get_collection/invalid_datatype_type.json"
    )
    res = client.get_collection("COLLECTION_ID")
    assert res["DATA_TYPE"] == "result#1.0.0"
    assert res.full_data["DATA_TYPE"] == "result#1.0.0"
    assert "detail" in res.full_data
    assert "detail" in res.data
    assert "data" in res.full_data
    assert res.full_data["detail"] == "success"


def test_delete_collection(client):
    register_api_route_fixture_file(
        "gcs", "/collections/COLLECTION_ID", "empty_success.json", method="DELETE"
    )
    res = client.delete_collection("COLLECTION_ID")
    assert res["DATA_TYPE"] == "result#1.0.0"
    assert "detail" in res.data
    assert res.data["detail"] == "success"


def test_create_mapped_collection(client):
    register_api_route_fixture_file(
        "gcs", "/collections", "create_collection.json", method="POST"
    )
    collection = MappedCollectionDocument(
        domain_name="i-f3c83.123.globus.org",
        display_name="Project Foo Research Data",
        identity_id="c8b7ab5c-595c-43c9-8e43-9e8a3debfe4c",
        storage_gateway_id="fc1f3ba0-1fa4-42b2-8bb3-53983774fa5f",
        collection_base_path="/",
        default_directory="/projects",
        public=True,
        force_encryption=False,
        disable_verify=False,
        organization="University of Example",
        department="Data Science",
        keywords=["Project Foo", "Data Intensive Science"],
        description='Information related to the "Foo" project.',
        contact_email="project-foo@example.edu",
        contact_info="+1 (555) 555-1234",
        info_link="https://project-foo.example.edu/info",
        policies={"DATA_TYPE": "blackpearl_collection_policies#1.0.0"},
        allow_guest_collections=True,
        sharing_restrict_paths={
            "DATA_TYPE": "path_restrictions#1.0.0",
            "read": ["/public"],
            "read_write": ["/home", "/projects"],
            "none": ["/private"],
        },
    )
    res = client.create_collection(collection)
    assert res["DATA_TYPE"] == "collection#1.0.0"
    assert res["display_name"] == "Project Foo Research Data"


def test_update_guest_collection(client):
    register_api_route_fixture_file(
        "gcs", "/collections/COLLECTION_ID", "update_collection.json", method="PATCH"
    )
    collection = GuestCollectionDocument(display_name="Project Foo Research Data")
    res = client.update_collection("COLLECTION_ID", collection)
    assert res["DATA_TYPE"] == "collection#1.0.0"
    assert res["display_name"] == "Project Foo Research Data"



from __future__ import annotations

import json

import pytest

import globus_sdk
from globus_sdk import FlowsAPIError, SpecificFlowClient
from globus_sdk.testing import get_last_request, load_response


def test_run_flow(specific_flow_client_class: type[SpecificFlowClient]):
    metadata = load_response(SpecificFlowClient.run_flow).metadata

    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    resp = flow_client.run_flow(**metadata["request_params"])
    assert resp.http_status == 200


def test_run_flow_missing_scope(specific_flow_client_class: type[SpecificFlowClient]):
    metadata = load_response(
        SpecificFlowClient.run_flow, case="missing_scope_error"
    ).metadata

    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    with pytest.raises(FlowsAPIError) as excinfo:
        flow_client.run_flow(**metadata["request_params"])

    err = excinfo.value
    assert err.http_status == 403
    assert err.code == "MISSING_SCOPE"
    assert (
        err.message == "This action requires the following scope: frobulate[demuddle]"
    )


def test_run_flow_without_activity_notification_policy(
    specific_flow_client_class,
):
    metadata = load_response(SpecificFlowClient.run_flow).metadata
    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])
    resp = flow_client.run_flow({})
    assert resp.http_status == 200

    last_req = get_last_request()
    sent_payload = json.loads(last_req.body)
    assert "activity_notification_policy" not in sent_payload


def test_run_flow_with_empty_activity_notification_policy(
    specific_flow_client_class,
):
    metadata = load_response(SpecificFlowClient.run_flow).metadata
    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    policy = globus_sdk.RunActivityNotificationPolicy()
    flow_client.run_flow({}, activity_notification_policy=policy)

    last_req = get_last_request()
    sent_payload = json.loads(last_req.body)
    assert "activity_notification_policy" in sent_payload
    assert sent_payload["activity_notification_policy"] == {}


def test_run_flow_with_activity_notification_policy(
    specific_flow_client_class,
):
    metadata = load_response(SpecificFlowClient.run_flow).metadata
    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    policy = globus_sdk.RunActivityNotificationPolicy(status=["FAILED", "INACTIVE"])
    flow_client.run_flow({}, activity_notification_policy=policy)

    last_req = get_last_request()
    sent_payload = json.loads(last_req.body)
    assert "activity_notification_policy" in sent_payload
    assert sent_payload["activity_notification_policy"] == {
        "status": ["FAILED", "INACTIVE"]
    }



#!/usr/bin/env python

import globus_sdk
from globus_sdk.experimental.globus_app import UserApp

# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"
USER_APP = UserApp("manage-timers-example", client_id=NATIVE_CLIENT_ID)


def main():
    client = globus_sdk.TimersClient(app=USER_APP)

    first = True
    for record in client.list_jobs(query_params={"filter_active": True})["jobs"]:
        if not first:
            print("---")
        first = False
        print("name:", record["name"])
        print("id:", record["job_id"])


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import pathlib
import typing as t


class LocalGlobusConnectServer:
    r"""
    A LocalGlobusConnectServer object represents the available SDK methods
    for inspecting and controlling a running Globus Connect Server
    installation.

    These objects do *not* inherit from BaseClient and do not provide methods
    for interacting with any Globus Service APIs.

    :param info_path: The path to the info file used to inspect the local system
    """

    def __init__(
        self,
        *,
        info_path: str | pathlib.Path = "/var/lib/globus-connect-server/info.json",
    ) -> None:
        self.info_path = pathlib.Path(info_path)
        self._loaded_info: dict[str, t.Any] | None = None

    @property
    def info_dict(self) -> dict[str, t.Any] | None:
        """
        The info.json data for the local Globus Connect Server, as a dict.

        If the info.json file is not present or cannot be parsed, the data is None.
        This indicates that there is no local Globus Connect Server node, or if there is
        one, it cannot be used or examined by the SDK. For example, containerized
        applications using the SDK may not be able to interact with Globus Connect
        Server on the container host.
        """
        if self._loaded_info is None:
            if self.info_path.is_file():
                with open(self.info_path, encoding="utf-8") as fp:
                    try:
                        parsed_data = json.load(fp)
                    except (UnicodeDecodeError, json.JSONDecodeError):
                        pass
                    else:
                        if isinstance(parsed_data, dict):
                            self._loaded_info = t.cast(t.Dict[str, t.Any], parsed_data)
        return self._loaded_info

    @info_dict.deleter
    def info_dict(self) -> None:
        self._loaded_info = None

    @property
    def endpoint_id(self) -> str | None:
        """
        The endpoint ID of the local Globus Connect Server endpoint.
        None if the data cannot be loaded or is malformed.
        """
        if self.info_dict is not None:
            epid = self.info_dict.get("endpoint_id")
            if isinstance(epid, str):
                return epid
        return None

    @property
    def domain_name(self) -> str | None:
        """
        The domain name of the local Globus Connect Server endpoint.
        None if the data cannot be loaded or is malformed.
        """
        if self.info_dict is not None:
            domain = self.info_dict.get("domain_name")
            if isinstance(domain, str):
                return domain
        return None



import textwrap

import pytest

import globus_sdk


def test_locate_class_finds_transfer_client(sphinxext):
    assert (
        sphinxext.utils.locate_class("globus_sdk.TransferClient")
        is globus_sdk.TransferClient
    )


def test_locate_class_rejects_missing(sphinxext):
    with pytest.raises(RuntimeError, match="MISSING is not a class name"):
        sphinxext.utils.locate_class("globus_sdk.MISSING")


def test_read_sphinx_params(sphinxext):
    docstring = """
    preamble

    :param param1: some doc on one line
    :param param2: other doc
        spanning multiple lines
    :param param3: a doc
        spanning
        many
        lines
    :param param4: a doc
        spanning lines

        with a break in the middle ^
    :param param5: another


    :param param6: and a final one after some whitespace

    epilogue
    """
    params = sphinxext.utils.read_sphinx_params(docstring)
    assert len(params) == 6
    assert params[0] == ":param param1: some doc on one line"
    assert params[1] == ":param param2: other doc\n    spanning multiple lines"
    assert params[2] == textwrap.dedent("""\
        :param param3: a doc
            spanning
            many
            lines""")
    assert params[3] == textwrap.dedent("""\
        :param param4: a doc
            spanning lines

            with a break in the middle ^""")
    assert params[4] == ":param param5: another"
    assert params[5] == ":param param6: and a final one after some whitespace"

    # clear the cache after the test
    # not essential, but reduces the risk that this impacts some future test
    sphinxext.utils.read_sphinx_params.cache_clear()



import time
from unittest import mock

from globus_sdk.token_storage.legacy import MemoryAdapter


def test_memory_adapter_store_overwrites_only_new_data():
    # setup a mock token response
    expiration_time = int(time.time()) + 3600
    mock_response = mock.Mock()
    mock_response.by_resource_server = {
        "resource_server_1": {
            "access_token": "access_token_1",
            "expires_at_seconds": expiration_time,
            "refresh_token": "refresh_token_1",
            "resource_server": "resource_server_1",
            "scope": "scope1",
            "token_type": "bearer",
        },
        "resource_server_2": {
            "access_token": "access_token_2",
            "expires_in": expiration_time,
            "refresh_token": "refresh_token_2",
            "resource_server": "resource_server_2",
            "scope": "scope2 scope2:0 scope2:1",
            "token_type": "bearer",
        },
    }

    # "store" it in memory
    adapter = MemoryAdapter()
    adapter.store(mock_response)

    # read back a sample piece of data
    fetch1 = adapter.get_token_data("resource_server_1")
    assert fetch1["access_token"] == "access_token_1"

    # "store" a new mock response which overwrites only one piece of data
    mock_response2 = mock.Mock()
    mock_response2.by_resource_server = {
        "resource_server_1": {
            "access_token": "access_token_1_new",
            "expires_at_seconds": expiration_time,
            "refresh_token": "refresh_token_1",
            "resource_server": "resource_server_1",
            "scope": "scope1",
            "token_type": "bearer",
        }
    }
    adapter.store(mock_response2)

    # the overwritten data is updated
    fetch2 = adapter.get_token_data("resource_server_1")
    assert fetch2["access_token"] == "access_token_1_new"

    # but the existing data is preserved
    fetch3 = adapter.get_token_data("resource_server_2")
    assert fetch3["access_token"] == "access_token_2"



from .client import TransferClient
from .data import CreateTunnelData, DeleteData, TransferData
from .errors import TransferAPIError
from .response import IterableTransferResponse

__all__ = (
    "TransferClient",
    "TransferData",
    "DeleteData",
    "TransferAPIError",
    "IterableTransferResponse",
    "CreateTunnelData",
)



from __future__ import annotations

import abc

import globus_sdk
from globus_sdk._missing import none2missing
from globus_sdk.gare import GlobusAuthorizationParameters


class LoginFlowManager(metaclass=abc.ABCMeta):
    """
    The abstract base class defining the interface for managing login flows.

    Implementing classes must supply a ``run_login_flow`` method.
    Utility functions starting an authorization-code grant flow and getting an
    authorization-code URL are provided on the class.

    :ivar AuthLoginClient login_client: A native or confidential login client to be
        used by the login flow manager.
    :ivar bool request_refresh_tokens: A signal of whether refresh tokens are expected
        to be requested, in addition to access tokens.
    :ivar str native_prefill_named_grant: A string to prefill in a Native App login
        flow. This value is only to be used if the `login_client` is a native client.
    """

    def __init__(
        self,
        login_client: globus_sdk.AuthLoginClient,
        *,
        request_refresh_tokens: bool = False,
        native_prefill_named_grant: str | None = None,
    ) -> None:
        if not isinstance(
            login_client,
            (globus_sdk.NativeAppAuthClient, globus_sdk.ConfidentialAppAuthClient),
        ):
            raise globus_sdk.GlobusSDKUsageError(
                f"{type(self).__name__} requires a NativeAppAuthClient or "
                f"ConfidentialAppAuthClient, but got a {type(login_client).__name__}."
            )

        self.login_client = login_client
        self.request_refresh_tokens = request_refresh_tokens
        self.native_prefill_named_grant = native_prefill_named_grant

    def _get_authorize_url(
        self, auth_parameters: GlobusAuthorizationParameters, redirect_uri: str
    ) -> str:
        """
        Utility method to provide a simpler interface for subclasses to start an
        authorization flow and get an authorization URL.
        """
        self._oauth2_start_flow(auth_parameters, redirect_uri)

        # prompt is assigned first because its usage is type-ignored below
        # this makes it clear that the ignore applies to that usage
        prompt = none2missing(auth_parameters.prompt)
        return self.login_client.oauth2_get_authorize_url(
            session_required_identities=none2missing(
                auth_parameters.session_required_identities
            ),
            session_required_single_domain=none2missing(
                auth_parameters.session_required_single_domain
            ),
            session_required_policies=none2missing(
                auth_parameters.session_required_policies
            ),
            session_required_mfa=none2missing(auth_parameters.session_required_mfa),
            session_message=none2missing(auth_parameters.session_message),
            prompt=prompt,  # type: ignore[arg-type]
        )

    def _oauth2_start_flow(
        self, auth_parameters: GlobusAuthorizationParameters, redirect_uri: str
    ) -> None:
        """
        Start an authorization flow with the class's login_client, returning an
        authorization URL to direct a user to.
        """
        login_client = self.login_client
        requested_scopes = auth_parameters.required_scopes
        if not requested_scopes:
            raise globus_sdk.GlobusSDKUsageError(
                f"{type(self).__name__} cannot start a login flow without scopes "
                "in the authorization parameters."
            )

        # Native and Confidential App clients have different signatures for this method,
        # so they must be type checked & called independently.
        if isinstance(login_client, globus_sdk.NativeAppAuthClient):
            login_client.oauth2_start_flow(
                requested_scopes,
                redirect_uri=redirect_uri,
                refresh_tokens=self.request_refresh_tokens,
                prefill_named_grant=none2missing(self.native_prefill_named_grant),
            )
        elif isinstance(login_client, globus_sdk.ConfidentialAppAuthClient):
            login_client.oauth2_start_flow(
                redirect_uri,
                requested_scopes,
                refresh_tokens=self.request_refresh_tokens,
            )

    @abc.abstractmethod
    def run_login_flow(
        self,
        auth_parameters: GlobusAuthorizationParameters,
    ) -> globus_sdk.OAuthTokenResponse:
        """
        Run a login flow to get tokens for a user.

        :param auth_parameters: ``GlobusAuthorizationParameters`` passed through
            to the authentication flow to control how the user will authenticate.
        """



"""
Definition and loading of standard environment variables, plus a wrappers for loading
and parsing values.

This does not include service URL env vars (see environments.py for loading of those)
"""

from __future__ import annotations

import logging
import os
import pathlib
import typing as t

log = logging.getLogger(__name__)

ENVNAME_VAR = "GLOBUS_SDK_ENVIRONMENT"
HTTP_TIMEOUT_VAR = "GLOBUS_SDK_HTTP_TIMEOUT"
SSL_VERIFY_VAR = "GLOBUS_SDK_VERIFY_SSL"


def get_environment_name(inputenv: str | None = None) -> str:
    if inputenv is None:
        value = os.getenv(ENVNAME_VAR, "production")
    else:
        value = inputenv
    log.debug(f"get_environment_name() got value: {value}")
    return value


def get_ssl_verify(value: bool | str | pathlib.Path | None = None) -> bool | str:
    if value is None:
        value = os.getenv(SSL_VERIFY_VAR, "1")
    value = _ssl_verify_cast(value)
    log.debug(f"get_ssl_verify() got value: {value}")
    return value


def get_http_timeout(value: float | None = None) -> float | None:
    if value is not None:
        result: float = value
    else:
        var = os.getenv(HTTP_TIMEOUT_VAR, "60")

        # TODO: re-evaluate converting `""` to a default timeout -- it seems like this
        # should error instead, but the behavior has been established for a long time,
        # so will need proper deprecation
        if var == "":
            result = 60.0
        else:
            result = _float_cast(var)

    log.debug(f"get_http_timeout() got value: {result}")
    if result == -1.0:
        return None
    return result


def _ssl_verify_cast(value: t.Any) -> bool | str:
    if isinstance(value, bool):
        return value
    if not isinstance(value, (str, pathlib.Path)):
        msg = f"Value {value} of type {type(value)} cannot be used for SSL verification"
        raise ValueError(msg)
    if isinstance(value, str):
        if value.lower() in {"y", "yes", "t", "true", "on", "1"}:
            return True
        if value.lower() in {"n", "no", "f", "false", "off", "0"}:
            return False
        if os.path.isfile(value):
            return value
    if isinstance(value, pathlib.Path) and value.is_file():
        return str(value.absolute())
    raise ValueError(
        "SSL verification value must be a valid boolean value "
        f"or a path to a file that exists (got {value})"
    )


def _float_cast(value: str) -> float:
    try:
        return float(value)
    except ValueError as e:
        log.error(f'Value "{value}" can\'t cast to float')
        raise ValueError(f"Invalid config float: {value}") from e



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

endpoint_client_id = str(uuid.uuid4())
domain_name = "abc.xyz.data.globus.org"
gcs_manager_url = RegisteredResponse._url_map["gcs"]

RESPONSES = ResponseSet(
    metadata={
        "endpoint_client_id": endpoint_client_id,
        "domain_name": domain_name,
    },
    default=RegisteredResponse(
        service="gcs",
        path="/info",
        json={
            "DATA_TYPE": "result#1.1.0",
            "code": "success",
            "data": [
                {
                    "DATA_TYPE": "info#1.0.0",
                    "api_version": "1.29.0",
                    "client_id": endpoint_client_id,
                    "domain_name": domain_name,
                    "endpoint_id": endpoint_client_id,
                    "manager_version": "5.4.76-rc3",
                },
                {
                    "DATA_TYPE": "connector#1.1.0",
                    "display_name": "POSIX",
                    "id": "145812c8-decc-41f1-83cf-bb2a85a2a70b",
                    "is_baa": False,
                    "is_ha": False,
                },
            ],
        },
    ),
)



import pytest

import globus_sdk


@pytest.fixture
def make_oauth_token_response(make_response):
    """
    response with conveniently formatted names to help with iteration in tests
    """

    def f(client=None):
        return make_response(
            response_class=globus_sdk.services.auth.response.OAuthTokenResponse,
            json_body={
                "access_token": "access_token_1",
                "expires_in": 3600,
                "id_token": "id_token_value",
                "refresh_token": "refresh_token_1",
                "resource_server": "resource_server_1",
                "scope": "scope1",
                "state": "provided_by_client_to_prevent_replay_attacks",
                "token_type": "bearer",
                "other_tokens": [
                    {
                        "access_token": "access_token_2",
                        "expires_in": 3600,
                        "refresh_token": "refresh_token_2",
                        "resource_server": "resource_server_2",
                        "scope": "scope2 scope2:0 scope2:1",
                        "token_type": "bearer",
                    },
                    {
                        "access_token": "access_token_3",
                        "expires_in": 3600,
                        "refresh_token": "refresh_token_3",
                        "resource_server": "resource_server_3",
                        "scope": "scope3:0 scope3:1",
                        "token_type": "bearer",
                    },
                ],
            },
            client=client,
        )

    return f


@pytest.fixture
def make_oauth_dependent_token_response(make_response):
    """
    response with conveniently formatted names to help with iteration in tests
    """

    def f(client=None):
        return make_response(
            response_class=(
                globus_sdk.services.auth.response.OAuthDependentTokenResponse
            ),
            json_body=[
                {
                    "access_token": "access_token_4",
                    "expires_in": 3600,
                    "refresh_token": "refresh_token_4",
                    "resource_server": "resource_server_4",
                    "scope": "scope4",
                    "token_type": "bearer",
                },
                {
                    "access_token": "access_token_5",
                    "expires_in": 3600,
                    "refresh_token": "refresh_token_5",
                    "resource_server": "resource_server_5",
                    "scope": "scope5",
                    "token_type": "bearer",
                },
            ],
            client=client,
        )

    return f


@pytest.fixture
def oauth_token_response(make_oauth_token_response):
    return make_oauth_token_response()


@pytest.fixture
def oauth_dependent_token_response(make_oauth_dependent_token_response):
    return make_oauth_dependent_token_response()



from globus_sdk.testing import load_response


def test_delete_project(service_client):
    meta = load_response(service_client.delete_project).metadata

    project_id = meta["id"]
    res = service_client.delete_project(project_id)

    assert res["project"]["id"] == meta["id"]



import urllib.parse

import pytest

from globus_sdk import MISSING
from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize("filter_fulltext", [MISSING, "foo"])
@pytest.mark.parametrize("filter_roles", [MISSING, "bar"])
@pytest.mark.parametrize("orderby", [MISSING, "created_at ASC"])
def test_list_flows_simple(flows_client, filter_fulltext, filter_roles, orderby):
    meta = load_response(flows_client.list_flows).metadata

    add_kwargs = {}
    if filter_fulltext:
        add_kwargs["filter_fulltext"] = filter_fulltext
    if filter_roles:
        add_kwargs["filter_roles"] = filter_roles
    if orderby:
        add_kwargs["orderby"] = orderby

    res = flows_client.list_flows(**add_kwargs)

    assert res.http_status == 200
    # dict-like indexing
    assert meta["first_flow_id"] == res["flows"][0]["id"]
    # list conversion (using __iter__) and indexing
    assert meta["first_flow_id"] == list(res)[0]["id"]

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    expect_query_params = {
        k: [v]
        for k, v in (
            ("filter_fulltext", filter_fulltext),
            ("filter_roles", filter_roles),
            ("orderby", orderby),
        )
        if v is not MISSING
    }
    assert parsed_qs == expect_query_params


@pytest.mark.parametrize("by_pages", [True, False])
def test_list_flows_paginated(flows_client, by_pages):
    meta = load_response(flows_client.list_flows, case="paginated").metadata
    total_items = meta["total_items"]
    num_pages = meta["num_pages"]
    expect_markers = meta["expect_markers"]

    res = flows_client.paginated.list_flows()
    if by_pages:
        pages = list(res)
        assert len(pages) == num_pages
        for i, page in enumerate(pages):
            assert page["marker"] == expect_markers[i]
            if i < num_pages - 1:
                assert page["has_next_page"] is True
            else:
                assert page["has_next_page"] is False
    else:
        items = list(res.items())
        assert len(items) == total_items


@pytest.mark.parametrize(
    "orderby_style, orderby_value",
    [
        # single str -> single param
        ("str", "created_at ASC"),
        # sequence of strs (tuple, list) -> multi param
        ("seq", ["created_at ASC", "updated_at DESC"]),
        ("seq", ("created_at ASC", "updated_at DESC")),
        # more complex cases to handle within the test: generators and sets
        ("generator", ("created_at ASC", "updated_at DESC")),
        ("set", ("created_at ASC", "updated_at DESC")),
    ],
)
def test_list_flows_orderby_multi(flows_client, orderby_style, orderby_value):
    meta = load_response(flows_client.list_flows).metadata

    if orderby_style == "str":
        orderby = orderby_value
        expected_orderby_value = [orderby_value]
    elif orderby_style == "seq":
        orderby = orderby_value
        expected_orderby_value = list(orderby_value)
    elif orderby_style == "generator":
        orderby = (x for x in orderby_value)
        expected_orderby_value = list(orderby_value)
    elif orderby_style == "set":
        orderby = set(orderby_value)
        expected_orderby_value = set(orderby_value)
    else:
        raise NotImplementedError

    res = flows_client.list_flows(orderby=orderby)
    assert res.http_status == 200
    # check result correctness
    assert meta["first_flow_id"] == res["flows"][0]["id"]

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)

    # for the set case, handle arbitrary ordering by converting the value seen
    # back to a set
    if isinstance(expected_orderby_value, set):
        assert set(parsed_qs["orderby"]) == expected_orderby_value
    else:
        assert parsed_qs["orderby"] == expected_orderby_value


@pytest.mark.parametrize(
    "filter_roles, expected_filter_roles",
    [
        # empty list/tuple, list/tuple with empty string, and None do not send the param
        ((), None),
        ([], None),
        ([""], None),
        (("",), None),
        (MISSING, None),
        # single role as string
        ("foo", ["foo"]),
        # single role as list/tuple
        (["foo"], ["foo"]),
        (("foo",), ["foo"]),
        # multiple roles as comma-separated string
        ("foo,bar", ["foo,bar"]),
        # multiple roles as list/tuple
        (["foo", "bar"], ["foo,bar"]),
        (("foo", "bar"), ["foo,bar"]),
    ],
)
def test_list_flows_with_filter_roles_parameter(
    flows_client, filter_roles, expected_filter_roles
):
    load_response(flows_client.list_flows).metadata

    res = flows_client.list_flows(filter_roles=filter_roles)
    assert res.http_status == 200

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)

    # Only check filter_roles in parsed_qs if we expect a non-empty value
    if expected_filter_roles:
        assert parsed_qs["filter_roles"] == expected_filter_roles
    else:
        assert "filter_roles" not in parsed_qs



import json
import urllib.parse

import pytest

from globus_sdk import MISSING
from globus_sdk.testing import get_last_request, load_response

_OMIT = object()


@pytest.mark.parametrize("local_user", ("my-user", MISSING, _OMIT))
def test_operation_rename(client, local_user):
    meta = load_response(client.operation_rename).metadata
    endpoint_id = meta["endpoint_id"]

    if local_user is not _OMIT:
        res = client.operation_rename(
            endpoint_id=endpoint_id,
            oldpath="~/old-name",
            newpath="~/new-name",
            local_user=local_user,
            query_params={"foo": "bar"},
        )
    else:
        res = client.operation_rename(
            endpoint_id=endpoint_id,
            oldpath="~/old-name",
            newpath="~/new-name",
            query_params={"foo": "bar"},
        )
    assert res["DATA_TYPE"] == "result"
    assert res["code"] == "FileRenamed"

    req = get_last_request()
    body = json.loads(req.body)
    assert body["old_path"] == "~/old-name"
    assert body["new_path"] == "~/new-name"
    if local_user not in (_OMIT, MISSING):
        assert body["local_user"] == local_user
    else:
        assert "local_user" not in body
    query_params = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    assert query_params["foo"] == ["bar"]



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

identity_id = str(uuid.uuid4())
collection_ids = [str(uuid.uuid4()), str(uuid.uuid4())]
gateway_ids = [str(uuid.uuid4()), str(uuid.uuid4())]
display_names = ["Happy Fun Collection Name 1", "Happy Fun Collection Name 2"]

RESPONSES = ResponseSet(
    metadata={
        "identity_id": identity_id,
        "collection_ids": collection_ids,
        "gateway_ids": gateway_ids,
        "display_names": display_names,
    },
    default=RegisteredResponse(
        service="gcs",
        path="/collections",
        json={
            "DATA_TYPE": "result#1.0.0",
            "code": "success",
            "detail": "success",
            "http_response_code": 200,
            "data": [
                {
                    "DATA_TYPE": "collection#1.0.0",
                    "public": True,
                    "id": collection_ids[0],
                    "display_name": display_names[0],
                    "identity_id": identity_id,
                    "collection_type": "mapped",
                    "storage_gateway_id": gateway_ids[0],
                    "require_high_assurance": False,
                    "high_assurance": False,
                    "authentication_assurance_timeout": 15840,
                    "authentication_timeout_mins": 15840,
                },
                {
                    "DATA_TYPE": "collection#1.0.0",
                    "public": True,
                    "id": collection_ids[1],
                    "display_name": display_names[1],
                    "identity_id": identity_id,
                    "collection_type": "mapped",
                    "storage_gateway_id": gateway_ids[1],
                    "require_high_assurance": False,
                    "high_assurance": False,
                    "authentication_assurance_timeout": 15840,
                    "authentication_timeout_mins": 15840,
                },
            ],
        },
    ),
    forbidden=RegisteredResponse(
        service="gcs",
        path="/collections",
        status=403,
        json={
            "code": "permission_denied",
            "http_response_code": 403,
            "DATA_TYPE": "result#1.0.0",
            "detail": None,
            "message": "Could not list collections. Insufficient permissions",
            "data": [],
            "has_next_page": False,
            "marker": "",
        },
    ),
)



import uuid

import pytest

import globus_sdk
from globus_sdk.testing import get_last_request, load_response


class StringWrapper:
    """Simple test object to be a non-string obj wrapping a string"""

    def __init__(self, s) -> None:
        self.s = s

    def __str__(self):
        return self.s


def test_get_identities_unauthorized(service_client):
    data = load_response(service_client.get_identities, case="unauthorized")

    with pytest.raises(globus_sdk.AuthAPIError) as excinfo:
        service_client.get_identities(usernames="foobar@example.com")

    err = excinfo.value
    assert err.code == "UNAUTHORIZED"
    assert data.metadata["error_id"] in err.text
    assert err.raw_json == data.json


@pytest.mark.parametrize(
    "usernames",
    [
        "globus@globus.org",
        StringWrapper("globus@globus.org"),
        ["globus@globus.org"],
        ("globus@globus.org",),
    ],
)
def test_get_identities_success(usernames, service_client):
    data = load_response(service_client.get_identities)
    res = service_client.get_identities(usernames=usernames)

    assert [x["id"] for x in res] == [data.metadata["id"]]

    lastreq = get_last_request()
    assert lastreq.params == {
        "usernames": "globus@globus.org",
        "provision": "false",  # provision defaults to false
    }


def test_get_identities_provision_flag_defaults_to_false(service_client):
    load_response(service_client.get_identities)
    service_client.get_identities(usernames="globus@globus.org")
    lastreq = get_last_request()
    assert "provision" in lastreq.params
    assert lastreq.params["provision"] == "false"


@pytest.mark.parametrize("inval, outval", [(True, "true"), (False, "false")])
def test_get_identities_provision_flag_formatting(inval, outval, service_client):
    load_response(service_client.get_identities)
    service_client.get_identities(usernames="globus@globus.org", provision=inval)
    lastreq = get_last_request()
    assert "provision" in lastreq.params
    assert lastreq.params["provision"] == outval


@pytest.mark.parametrize(
    "usernames",
    [
        "globus@globus.org,sirosen@globus.org",
        (StringWrapper("sirosen@globus.org"), StringWrapper("globus@globus.org")),
        ["globus@globus.org", "sirosen@globus.org"],
    ],
)
def test_get_identities_multiple_usernames_success(usernames, service_client):
    data = load_response(service_client.get_identities, case="multiple")
    if isinstance(usernames, str):
        expect_param = usernames
    else:
        expect_param = ",".join([str(x) for x in usernames])

    res = service_client.get_identities(usernames=usernames)

    assert [x["username"] for x in res] == data.metadata["usernames"]
    assert [x["id"] for x in res] == data.metadata["ids"]

    lastreq = get_last_request()
    assert "usernames" in lastreq.params
    assert lastreq.params["usernames"] == expect_param


@pytest.mark.parametrize(
    "ids",
    [
        # two uuids, already comma delimited
        ",".join(
            [
                "46bd0f56-e24f-11e5-a510-131bef46955c",
                "ae341a98-d274-11e5-b888-dbae3a8ba545",
            ]
        ),
        # a list of two UUID objects
        [
            uuid.UUID("46bd0f56-e24f-11e5-a510-131bef46955c"),
            uuid.UUID("ae341a98-d274-11e5-b888-dbae3a8ba545"),
        ],
    ],
)
def test_get_identities_multiple_ids_success(ids, service_client):
    data = load_response(service_client.get_identities, case="multiple")
    expect_param = ",".join(data.metadata["ids"])

    res = service_client.get_identities(ids=ids)

    assert [x["id"] for x in res] == data.metadata["ids"]
    assert [x["username"] for x in res] == data.metadata["usernames"]

    lastreq = get_last_request()
    assert "ids" in lastreq.params
    assert lastreq.params["ids"] == expect_param



import uuid

from globus_sdk import AuthClient, Scope
from globus_sdk.scopes import ScopeParser

# setup: get a consent forest
ac = AuthClient()
identity_id = uuid.uuid1()
consents = ac.get_consents(identity_id)
consent_forest = consents.to_forest()

# create some variant types
xfer_str: str = "urn:globus:auth:scope:transfer.api.globus.org:all"
strlist: list[str] = [xfer_str]
scopelist: list[Scope] = ScopeParser.parse(xfer_str)
scopeobj: Scope = Scope.parse(xfer_str)

# all should be allowed
b: bool
b = consent_forest.meets_scope_requirements(xfer_str)
b = consent_forest.meets_scope_requirements((xfer_str,))
b = consent_forest.meets_scope_requirements(strlist)
b = consent_forest.meets_scope_requirements(scopelist)

# and we really are validating the type, since a bool or list[bool] is not allowed
consent_forest.meets_scope_requirements(b)  # type: ignore[arg-type]
blist: list[bool] = [b]
consent_forest.meets_scope_requirements(blist)  # type: ignore[arg-type]



import typing as t
import uuid

import pytest

import globus_sdk
from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "transform",
    (
        pytest.param(lambda x: x, id="string"),
        pytest.param(lambda x: [x], id="list"),
        pytest.param(lambda x: {x}, id="set"),
        pytest.param(lambda x: uuid.UUID(x), id="uuid"),
        pytest.param(lambda x: [uuid.UUID(x)], id="uuid_list"),
    ),
)
def test_get_task_batch(
    compute_client_v2: globus_sdk.ComputeClientV2, transform: t.Callable
):
    meta = load_response(compute_client_v2.get_task_batch).metadata
    task_ids = transform(meta["task_id"])
    res = compute_client_v2.get_task_batch(task_ids=task_ids)

    assert res.http_status == 200
    assert meta["task_id"] in res.data["results"]



import typing as t
import uuid

from responses.matchers import json_params_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

_COMMON_RESPONSE_RECORD: dict[str, t.Any] = {
    "fqdns": [],
    "links": {"privacy_policy": None, "terms_and_conditions": None},
    "parent_client": None,
    "preselect_idp": None,
    "prompt_for_named_grant": True,
    "redirect_uris": [],
    "required_idp": None,
    "scopes": [],
    "userinfo_from_effective_identity": True,
}

PUBLIC_CLIENT_RESPONSE_RECORD: dict[str, t.Any] = {
    "client_type": "public_installed_client",
    "grant_types": ["authorization_code", "refresh_token"],
    **_COMMON_RESPONSE_RECORD,
}


def register_response(
    args: t.Mapping[str, t.Any],
) -> RegisteredResponse:
    # Some name of args to create_client() have differenlty named fields.
    body_fields: t.Dict[str, t.Any] = {}
    for arg_name in args:
        if arg_name == "terms_and_conditions" or arg_name == "privacy_policy":
            body_fields["links"] = {
                arg_name: args[arg_name],
                **body_fields.get("links", {}),
            }
        else:
            body_fields[arg_name] = args[arg_name]

    client_id = str(uuid.uuid1())

    # Default to a public client response unless arg says otherwise
    client_response_record = {
        **PUBLIC_CLIENT_RESPONSE_RECORD,
        **body_fields,
        "id": client_id,
    }

    return RegisteredResponse(
        service="auth",
        method="PUT",
        path=f"/v2/api/clients/{client_id}",
        json={"client": client_response_record},
        metadata={
            # Test functions use 'args' to form request
            "args": {**args, "client_id": client_id},
            # Test functions use 'response' to verify response
            "response": body_fields,
        },
        match=[json_params_matcher({"client": body_fields})],
    )


RESPONSES = ResponseSet(
    default=register_response({}),
    name=register_response({"name": str(uuid.uuid4()).replace("-", "")}),
    publicly_visible=register_response({"visibility": "public"}),
    not_publicly_visible=register_response({"visibility": "private"}),
    redirect_uris=register_response({"redirect_uris": ["https://foo.com"]}),
    links=register_response(
        {
            "terms_and_conditions": "https://foo.org",
            "privacy_policy": "https://boo.org",
        }
    ),
    required_idp=register_response({"required_idp": str(uuid.uuid1())}),
    preselect_idp=register_response({"preselect_idp": str(uuid.uuid1())}),
)



from __future__ import annotations

import contextlib
import dataclasses
import random
import typing as t

from .retry import RetryCheckCollection, RetryContext


def _exponential_backoff(ctx: RetryContext) -> float:
    # respect any explicit backoff set on the context
    if ctx.backoff is not None:
        return ctx.backoff
    # exponential backoff with jitter
    return t.cast(float, (0.25 + 0.5 * random.random()) * (2**ctx.attempt))


@dataclasses.dataclass
class RetryConfig:
    """
    Configuration for a client which is going to retry requests.

    :param max_retries: The maximum number of retries allowed.
    :param max_sleep: The maximum sleep time between retries (in seconds). If the
        computed sleep time or the backoff requested by a retry check exceeds this
        value, this amount of time will be used instead.
    :param backoff: A function which determines how long to sleep between calls
        based on the RetryContext. Defaults to exponential backoff with jitter based on
        the context ``attempt`` number.
    :param retry_after_status_codes: HTTP status codes for responses which may have
        a Retry-After header.
    :param transient_error_status_codes: HTTP status codes for error responses which
        should generally be retried.
    :param expired_authorization_status_codes: HTTP status codes indicating that
        authorization info was missing or expired.
    :param checks: The check callbacks which will run in order to evaluate
        responses and exceptions, as a ``RetryCheckCollection``.
    """

    max_retries: int = 5
    max_sleep: float | int = 10
    backoff: t.Callable[[RetryContext], float] = _exponential_backoff
    retry_after_status_codes: tuple[int, ...] = (429, 503)
    transient_error_status_codes: tuple[int, ...] = (429, 500, 502, 503, 504)
    expired_authorization_status_codes: tuple[int, ...] = (401,)

    checks: RetryCheckCollection = dataclasses.field(
        default_factory=RetryCheckCollection
    )

    @contextlib.contextmanager
    def tune(
        self,
        *,
        backoff: t.Callable[[RetryContext], float] | None = None,
        max_sleep: float | int | None = None,
        max_retries: int | None = None,
    ) -> t.Iterator[None]:
        """
        Temporarily adjust some of the request retry settings.
        This method works as a context manager, and will reset settings to their
        original values after it exits.

        :param backoff: A function which determines how long to sleep between
            calls based on the RetryContext
        :param max_sleep: The maximum sleep time between retries (in seconds). If the
            computed sleep time or the backoff requested by a retry check exceeds this
            value, this amount of time will be used instead
        :param max_retries: The maximum number of retries allowed by this transport

        **Example Usage**

        This can be used with any client class to temporarily set values in the context
        of one or more HTTP requests. For example, to disable retries:

        >>> client = ...  # any client class
        >>> with client.retry_config.tune(max_retries=0):
        >>>     foo = client.get_foo()

        See also: :meth:`RequestsTransport.tune`.
        """
        saved_settings = (
            self.backoff,
            self.max_sleep,
            self.max_retries,
        )
        if backoff is not None:
            self.backoff = backoff
        if max_sleep is not None:
            self.max_sleep = max_sleep
        if max_retries is not None:
            self.max_retries = max_retries
        yield
        (
            self.backoff,
            self.max_sleep,
            self.max_retries,
        ) = saved_settings



from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    from ._model import Consent


class ConsentParseError(Exception):
    """An error raised if consent parsing/loading fails."""

    def __init__(self, message: str, raw_consent: dict[str, t.Any]) -> None:
        super().__init__(message)
        self.raw_consent = raw_consent


class ConsentTreeConstructionError(Exception):
    """An error raised if consent tree construction fails."""

    def __init__(self, message: str, consents: list[Consent]) -> None:
        super().__init__(message)
        self.consents = consents



import globus_sdk

# your client credentials
CLIENT_ID = "YOUR ID HERE"
CLIENT_SECRET = "YOUR SECRET HERE"

# the ID of "tutorial collection 1"
TUTORIAL_COLLECTION = "6c54cade-bde5-45c1-bdea-f4bd71dba2cc"

# create a ClientApp named "app"
with globus_sdk.ClientApp(
    "sample-app",
    client_id=CLIENT_ID,
    client_secret=CLIENT_SECRET,
    config=globus_sdk.GlobusAppConfig(token_storage="memory"),
) as app:
    # create a TransferClient named "tc", bound to "app"
    with globus_sdk.TransferClient(app=app) as tc:

        # because the tutorial collection is of a type which uses a `data_access`
        # scope for fine grained access control, the `data_access` requirement needs
        # to be registered with the app
        tc.add_app_data_access_scope(TUTORIAL_COLLECTION)

        # iterate over the listing results, printing each filename
        for entry in tc.operation_ls(TUTORIAL_COLLECTION, path="/home/share/godata"):
            print(entry["name"])



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="search",
        method="POST",
        path=f"/v1/index/{INDEX_ID}/reopen",
        json={
            "index_id": INDEX_ID,
            "acknowledged": True,
        },
        metadata={"index_id": INDEX_ID},
    ),
    already_open=RegisteredResponse(
        service="search",
        method="POST",
        path=f"/v1/index/{INDEX_ID}/reopen",
        status=409,
        json={
            "code": "Conflict.IncompatibleIndexStatus",
            "request_id": "e1ad6822156dea372027eee48c16e150",
            "@datatype": "GError",
            "message": (
                "Index status (open) did not match required status for "
                "this operation: delete_pending"
            ),
            "@version": "2017-09-01",
            "status": 409,
        },
        metadata={"index_id": INDEX_ID},
    ),
)
